"""
Data management
"""


from .dates import *
from .dates import __all__ as dates_all

from .databanks import *
from .databanks import __all__ as databanks_all

__all__ = ( 
    []
    + dates_all 
    + databanks_all
)




"""
Mixin classes for displaying data management objects
"""


#[
from __future__ import annotations

from numbers import (Number, )
import numpy as np_
import re as re_
#]


_VERTICAL_ELLIPSIS = "⋮"
_REPEAT_SHORT_ROW = 2
_REPR_MAX_LEN = 70
_REPR_CONT = "..."
_REPR_INDENT = "    "
_REPR_SEPARATOR = ": "


def _get_series_row_str_(date, data, date_str_format, numeric_format, missing_str: str):
    """
    String representing one row of time series data including the date
    """
    #[
    date_str = ("{:"+date_str_format+"}").format(date)
    value_format = "{:"+numeric_format+"}"
    data_str = "".join([value_format.format(v) for v in data])
    data_str = data_str.replace("nan", "{:>3}".format(missing_str))
    return date_str + data_str
    #]


def _databank_repr(x, /, ) -> str:
    """
    String representing one record in a databank
    """
    #[
    if hasattr(x, "_databank_repr"):
        s = x._databank_repr()
    elif x is None:
        s = "None"
    elif x is ...:
        s = "..."
    elif isinstance(x, Number):
        s = str(x)
    elif isinstance(x, str):
        s = f'"{x}"'
    elif isinstance(x, np_.ndarray) or isinstance(x, list) or isinstance(x, tuple):
        s = re_.sub("\n + ", " ", repr(x))
    elif hasattr(x, "_get_first_line_view"):
        s = x._get_first_line_view()
    else:
        s = repr(type(x))
    return s if len(s)<_REPR_MAX_LEN else s[0:_REPR_MAX_LEN] + _REPR_CONT
    #]


class ViewMixin:
    """
    """
    _short_rows_: int = 5
    #[
    def _get_header_view_(self, /, ):
        """
        """
        return [ 
            "", 
            self._get_first_line_view(),
            f"Description: \"{self.get_description()}\"",
            "", 
        ]

    def _get_footer_view_(self, /, ):
        return ["", ]

    def _get_view_(self, /, ):
        """
        """
        header_view = self._get_header_view_()
        content_view = self._get_content_view_()
        footer_view = self._get_footer_view_()
        return header_view + content_view + footer_view

    def __invert__(self):
        """
        ~self for short view
        """
        header_view = self._get_header_view_()
        content_view = self._get_content_view_()
        if len(content_view) > 2*self._short_rows_:
            content_view = content_view[:self._short_rows_] + [self._get_short_row_()]*_REPEAT_SHORT_ROW + content_view[-self._short_rows_:]
        print("\n".join(header_view + content_view))

    def __repr__(self, /, ):
        """
        """
        return "\n".join(self._get_view_())

    def __str__(self, /, ):
        """
        """
        return repr(self)
    #]


class SeriesViewMixin(ViewMixin):
    """
    """
    #[
    def _get_first_line_view(self, /, ):
        """
        """
        shape = self.data.shape
        return f"Series {self.frequency.letter} {self.start_date}:{self.end_date} {shape[0]}-by-{shape[1]}"

    def _get_content_view_(self, /, ):
        """
        """
        return [
            _get_series_row_str_(date, data_row, self._date_str_format, self._numeric_format, self._missing_str) 
            for row, (date, data_row) in enumerate(zip(self.range, self.data))
        ]

    def _get_short_row_(self):
        """
        Create string representing rows skipped in the short view of times series
        """
        shape = self.data.shape
        date_str = ("{:"+self._date_str_format+"}").format(_VERTICAL_ELLIPSIS)
        value_format = "{:"+self._short_str_format+"}"
        data_str = "".join([value_format.format(_VERTICAL_ELLIPSIS)] * shape[1])
        return date_str + data_str
    #]


class DatabankViewMixin(ViewMixin):
    """
    """
    #[
    def _get_first_line_view(self, /, ):
        """
        """
        return f"Databank with {self._get_num_records():g} record(s)"

    def _get_content_view_(self, /, ):
        """
        """
        names = self._get_names()
        content = []
        if names:
            max_len = self._get_max_name_length_()
            content = [ _REPR_INDENT + str(k).rjust(max_len) + _REPR_SEPARATOR + _databank_repr(getattr(self, k)) for k in names ]
            content = content
        return content

    def _get_max_name_length_(self, /, ):
        """
        """
        return max(len(str(k)) for k in self._get_names())

    def _get_short_row_(self):
        """
        """
        max_len = self._get_max_name_length_()
        return _REPR_INDENT + _VERTICAL_ELLIPSIS.rjust(max_len) + " "*len(_REPR_SEPARATOR) + _VERTICAL_ELLIPSIS #]




"""
"""


#[
import numpy as _np
from typing import (Callable, )
from collections.abc import (Iterable, )

from .. import (equations as eq_, )
from . import (abc as _abc, )
#]


class PlainEquator(_abc.EquatorMixin, ):
    """
    """
    #[
    def __init__(
        self,
        equations: eq_.Equations,
        /,
        *,
        custom_functions: dict[str, Callable] | None = None,
    ) -> None:
        self._equations = tuple(equations, )
        self._create_equator_function(custom_functions, )
        self._populate_min_max_shifts()

    @property
    def min_num_columns(self, /, ) -> int:
        return -self.min_shift + 1 + self.max_shift

    def eval(
        self,
        data_array: _np.ndarray,
        columns: int | Iterable[int],
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        shape = (self.num_equations, -1)
        return self._func(data_array, columns, steady_array, ).reshape(shape)
    #]





"""
"""


#[
from __future__ import annotations

import re as re_
import parsimonious
import functools
import itertools
from collections.abc import Iterable
from typing import (TypeAlias, Protocol, )

from ..parsers import (common as co_, pseudofunctions as pf_, shifts as sh_, )
#]


def from_string(
    source: str,
    context: dict | None = None,
    /,
    save_preparsed: str = "",
) -> tuple[str, dict]:
    """
    """
    info = {
        "context": None,
        "preparser_needed": None,
    }

    info["context"] = context if context else {}

    # Remove line comments %, #, ...
    source = _remove_comments(source)

    # Evaluate <...> expressions in the local context
    source = _evaluate_contextual_expressions(source, info["context"])

    # Replace time shifts {...} by [...]
    source = sh_.standardize_time_shifts(source)

    # Replace human prefix (!) with processing prefix (¡)
    source = _translate_keywords(source)

    # Add blank lines pre and post source
    source = co_.add_blank_lines(source)

    info["preparser_needed"] = _is_preparser_needed(source)

    preparsed_source = _run_preparser_on_source_string(source, context, ) if info["preparser_needed"] else source

    # Resolve pseudofunctions
    preparsed_source = pf_.resolve_pseudofunctions(preparsed_source)

    if save_preparsed:
        _save_preparsed_source(preparsed_source, save_preparsed)

    return preparsed_source, info


_GRAMMAR_DEF = co_.GRAMMAR_DEF + r"""

    source =  (control / text)+
    control = white_spaces (for_do_block / if_then_block / else_keyword / end_keyword)
    text = ~r"[^¡]+"

    keyword = (
        for_keyword / do_keyword / 
        if_keyword / then_keyword / else_keyword /
        end_keyword
    )

    for_do_block = for_keyword white_spaces for_control do_keyword
        for_control = for_control_name_equals? for_token_ended*
        for_control_name_equals = for_control_name white_spaces "=" white_spaces
        for_control_name = ~r"\?\w*"
        for_token_ended = for_token for_token_end
        for_token = ~r"\w+"
        for_token_end = ~r"[,;\s]+"
        for_keyword = keyword_prefix "for"
        do_keyword = keyword_prefix "do"

    if_then_block = if_keyword if_condition then_keyword
        if_condition = ~r"[^¡]+"
        if_keyword = keyword_prefix "if"
        then_keyword = keyword_prefix "then"
        else_keyword = keyword_prefix "else"

    end_keyword = keyword_prefix "end"

"""


_GRAMMAR = parsimonious.grammar.Grammar(_GRAMMAR_DEF)
_KEYWORD_PREFIX = _GRAMMAR["keyword_prefix"].literal
_KEYWORDS = [ k.members[1].literal for k in _GRAMMAR["keyword"].members ]
_KEYWORDS_PATTERN = co_.compile_keywords_pattern(_KEYWORDS)
_translate_keywords = functools.partial(co_.translate_keywords, _KEYWORDS_PATTERN)


class _Visitor(parsimonious.nodes.NodeVisitor):
    """
    """
    #[
    def __init__(self, context=None):
        super().__init__()
        self.content = []
        self.context = context if context else {}

    def _add(self, new):
        self.content.append(new)

    def visit_source(self, node, visited_children):
        return self.content

    def visit_end_keyword(self, node, visited_children):
        self._add(_End())

    def visit_for_do_block(self, node, visited_children):
        control_name, tokens = visited_children[2]
        self._add(_For(control_name, tokens))

    def visit_for_control(self, node, visited_children):
        control_name = visited_children[0][0] if visited_children[0] else "?"
        tokens = visited_children[1]
        return [control_name, tokens]

    def visit_for_token_ended(self, node, visited_children):
        return visited_children[0]

    def visit_for_control_name_equals(self, node, visited_children):
        return visited_children[0]

    def visit_if_then_block(self, node, visited_children):
        condition = visited_children[1].strip()
        self._add(_If(condition))

    def visit_text(self, node, visited_children):
        text = _strip_lines(node.text)
        if text:
            self._add(_Text(text))

    def generic_visit(self, node, visited_children):
        return visited_children or node.text
    #]


class _ElementProtocol(Protocol, ):
    def resolve(self, sequence, context, ): ...
    def replace(self, pattern, replacement, ): ...


Sequence: TypeAlias = Iterable[_ElementProtocol]


class _Text:
    """
    """
    #[
    level = 0
    def __init__(self, content: str, ):
        self.content = content

    def replace(self, pattern: str, replacement: str, ):
        new = _Text(self.content.replace(pattern, replacement))
        return new

    def resolve(self, sequence: Sequence, context: dict, ):
        return self.content, sequence[1:]
    #]


class _For:
    """
    """
    #[
    level = 1
    def __init__(self, control_name, tokens, /, ):
        self.control_name = control_name
        self.tokens = tokens

    def replace(self, pattern, replacement, /, ):
        return self

    def _resolve_sequence(self, sequence, context, control_name, token, /, ):
        sequence = [ s.replace(control_name, token) for s in sequence ]
        return _resolve_sequence(sequence, context)

    def resolve(self, sequence: Sequence, context: dict, /, ):
        index_end = _find_matching_end(sequence)
        code = "\n".join(
            self._resolve_sequence(sequence[1:index_end], context, self.control_name, t)
            for t in self.tokens
        )
        return code, sequence[index_end+1:]
    #]


class _If:
    """
    """
    #[
    level = 1

    def __init__(self, condition, /, ):
        self.condition_text = condition
        self.condition_result = None

    def replace(self, pattern, replacement):
        self.condition_text = self.condition_text.replace(pattern, replacement)
        return self

    def resolve(self, sequence: Sequence, context: dict, /, ):
        raise Exception("!if not implemented yet")
        index_end = _find_matching_end(sequence)
        code = ...
        return code, sequence[index_end+1:]
    #]

class _End:
    """
    """
    #[
    level = -1

    def replace(self, pattern, replacement):
        return self
    #]


_COMMENTS_PATTERN = re_.compile(r'"[^"\n]*"|[%#].*|\.\.\..*')


def _remove_comments(source: str, /, ) -> str:
    return re_.sub(
        _COMMENTS_PATTERN,
        lambda m: m.group(0) if m.group(0).startswith('"') else "",
        source,
    )


def _find_matching_end(sequence: Sequence, /, ) -> int:
    level = list(itertools.accumulate(s.level for s in sequence))
    return level.index(0)


def _resolve_sequence(sequence: Sequence, context: dict, /, ) -> str:
    code = ""
    while sequence:
        new_code, sequence = sequence[0].resolve(sequence, context, )
        if new_code:
            code = code + "\n" + new_code if code else new_code
    return code


def _strip_lines(text: str) -> str:
    split_text = (t.strip() for t in text.split("\n"))
    return "\n".join(s for s in split_text if s)


_contextual_expression_pattern = re_.compile(r"<([^>]*)>")


def _stringify(input, /, ):
    if isinstance(input, str):
        return input
    elif not isinstance(input, Iterable):
        return str(input)
    else:
        return ",".join([_stringify(i, ) for i in input])

def _evaluate_contextual_expressions(text: str, context: dict, /):
    def _replace(match):
        try:
            return _stringify(eval(match.group(1), {}, context), )
        except:
            raise Exception(f"Error evaluating this contextual expresssion: {match.group(0)}")
    return re_.sub(_contextual_expression_pattern, _replace, text)


def _is_preparser_needed(source: str, /, ) -> bool:
    return _KEYWORD_PREFIX in source


def _save_preparsed_source(source: str, file_name: str, /, ) -> None:
    if file_name:
        with open(file_name, "wt+") as fid:
            fid.write(source)

def _run_preparser_on_source_string(source: str, context: dir, /, ) -> str:
    nodes = _GRAMMAR["source"].parse(source, )
    visitor = _Visitor()
    sequence = visitor.visit(nodes, )
    return _resolve_sequence(sequence, context, )





"""
Time series
"""


from .facade import *
from .facade import __all__ as facade_all

from .plotly import *
from .plotly import __all__ as plotly_all

__all__ = ( 
    []
    + facade_all
    + plotly_all
)




"""
General mixins
"""

#[
#]


class DescriptionMixin:
    """
    """
    #[
    def set_description(self, description: str, /, ) -> None:
        self._description = str(description)

    def get_description(self, /, ) -> str:
        return str(self._description)
    #]




"""Handle wrongdoings"""


#[
import warnings as wa_
from typing import (TypeAlias, Literal, )
from collections.abc import (Iterable, )
#]


HOW: TypeAlias = Literal["error"] | Literal["warning"] | Literal["silent"]

_PLAIN_PREFIX = ""
_LIST_PREFIX = "××× "


class IrisPieError(Exception):
    """
    """
    #[
    def __init__(self, message, ):
        message = _prepare_message(message)
        super().__init__(message)
    #]


class IrisPieWarning(UserWarning):
    """
    """
    pass


def throw(
    how: HOW,
    message: str | Iterable[str],
) -> None:
    """
    """
    #[
    _RESOLVE_HOW[how](message)
    #]


def _prepare_message(message):
    #[
    if isinstance(message, str):
        message = _PLAIN_PREFIX + message
    else:
        message = ("\n"+_LIST_PREFIX).join(message)
    return message
    #]


def _throw_as_error(
    message: str | Iterable[str],
) -> None:
    """
    """
    #[
    raise IrisPieError(message)
    #]


def _throw_as_warning(
    message: str | Iterable[str],
) -> None:
    """
    """
    #[
    message = _prepare_message(message)
    message = "\nIrisPieWarning: " + message
    wa_.warn(message, IrisPieWarning, stacklevel=4, )
    #]


def _throw_as_silent(
    message: str | Iterable[str],
) -> None:
    """
    """
    #[
    pass
    #]


_RESOLVE_HOW = {
    "error": _throw_as_error,
    "warning": _throw_as_warning,
    "silent": _throw_as_silent,
}






from __future__ import annotations

import re
import parsimonious
from collections.abc import Iterable


GRAMMAR_DEF = r"""

    keyword_prefix = "¡"
    human_prefix = "!"
    white_spaces = ~r"[ \t\n]*"
    non_keyword_prefixes = ~r"[^¡]+"

"""


GRAMMAR = parsimonious.grammar.Grammar(GRAMMAR_DEF)
KEYWORD_PREFIX = GRAMMAR["keyword_prefix"].literal
HUMAN_PREFIX = GRAMMAR["human_prefix"].literal


def compile_keywords_pattern(commands_list: Iterable[str]) -> re.Pattern:
    return re.compile(
        HUMAN_PREFIX + r"(" + "|".join(commands_list) + r")\b"
    )


def translate_keywords(commands_pattern: re.Pattern, code: str) -> str:
    return re.sub(commands_pattern, lambda m: KEYWORD_PREFIX + m.group(1), code)


def add_blank_lines(source: str) -> str:
    return "\n" + source + "\n"


def combine_source_files(
    source_files: str|Iterable[str], 
    /,
    joint: str = "\n\n",
) -> str:
    """
    """
    if isinstance(source_files, str):
        source_files = [source_files]
    source_string = []
    for f in source_files:
        with open(f, "r") as fid:
            source_string.append(fid.read())
    source_string = "\n\n".join(source_string)
    return source_string





"""
"""

#[
import numpy as _np
import scipy as _sp
from typing import (Protocol, Callable, )
from collections.abc import (Iterable, )

from ..aldi import (differentiators as _ad, maps as _am, )
from .. import (equations as _eq, quantities as _qu, incidence as _in, )
#]


class Jacobian:
    """
    # Describe the Jacobian matrix of a system of equations:
    * _shape -- shape of the Jacobian matrix (num_rows, num_columns)
    * _map -- mapping from (row, column) to (equation, quantity)
    * _aldi_context -- context for algorithmic differentiator wrt levels
    * _change_aldi_context -- context for algorithmic differentiator wrt changes
    * _array_creator -- function to create a dense or sparse Jacobian matrix
    * is_sparse -- whether the Jacobian matrix is sparse
    """
    #[
    atom_factory: _ad.AtomFactoryProtocol | None = None

    def __init__(
        self, 
        equations: _eq.Equations,
        wrt_qids: Iterable[int],
        qid_to_logly: dict[int, bool],
        /,
        *,
        custom_functions: dict[str, Callable] | None = None,
        **kwargs,
    ) -> None:
        """
        """
        #
        # Choose dense or sparse array creator
        self.is_sparse = kwargs.get("sparse_jacobian", False)
        self._create_jacobian = (
            _create_sparse_jacobian
            if self.is_sparse else _create_dense_jacobian
        )
        #
        # Extract eids from equations
        eids = tuple(eqn.id for eqn in equations)
        #
        # Collect the qids w.r.t. which each equation is to be differentiated
        wrt_qids = tuple(wrt_qids)
        eid_to_wrt_qids = {
            eqn.id: list(_generate_flat_wrt_qids_in_equation(eqn, wrt_qids, ))
            for eqn in equations
        }
        #
        # Create the map from eids to rhs offsets; the offset is the number
        # of rows in the Jacobian matrix that precede the equation
        eid_to_rhs_offset = _am.create_eid_to_rhs_offset(eids, eid_to_wrt_qids, )
        #
        self._shape = len(eids), len(wrt_qids),
        #
        self._map = _am.ArrayMap.for_equations(
            eids,
            eid_to_wrt_qids,
            wrt_qids,
            eid_to_rhs_offset,
            rhs_column=0,
            lhs_column_offset=0,
        )
        #
        self._aldi_context = _ad.Context(
            equations,
            self.atom_factory,
            eid_to_wrts=eid_to_wrt_qids,
            qid_to_logly=qid_to_logly,
            num_columns_to_eval=1,
            custom_functions=custom_functions,
        )


def _create_dense_jacobian(shape, diff_array, map, /, ) -> _np.ndarray:
    """
    Create Jacobian as numpy array
    """
    J = _np.zeros(shape, dtype=float, )
    J[map.lhs] = diff_array[map.rhs]
    return J


def _create_sparse_jacobian(shape, diff_array, map, /, ) -> _sp.sparse.coo_matrix:
    """
    Create Jacobian as scipy sparse matrix
    """
    J = _sp.sparse.coo_matrix(
        (diff_array[map.rhs], (map.lhs[0], map.lhs[1], )),
        shape, dtype=float,
    )
    return J


def _generate_flat_wrt_qids_in_equation(equation, all_wrt_qids):
    """
    Generate subset of the wrt_qids that occur in this equation (no matter what shift)
    """
    return ( 
        qid for qid in all_wrt_qids 
        if _in.is_qid_in_tokens(equation.incidence, qid)
    )





"""
Time series facade
"""


#[
from __future__ import annotations

from numbers import Number
from collections.abc import (Iterable, Callable, )
from typing import (Self, TypeAlias, )
from types import (EllipsisType, )
import numpy as np_
import itertools as it_
import copy as co_

from ..user import views as _vi
from ..user import description as _ud
from ..dataman import dates as _da
from . import filters as _sg
from . import plotly as _sp
from . import conversion as _sc
from . import x13 as _sx
#]


FUNCTION_ADAPTATIONS = ["log", "exp", "sqrt", "maximum", "minimum"] + ["cumsum"] + ["maxi", "mini", "mean", "median"]
np_.maxi = np_.max
np_.mini = np_.min


__all__ = [
    "Series", "shift", "diff", "difflog", "pct", "roc",
] + FUNCTION_ADAPTATIONS


for n in FUNCTION_ADAPTATIONS:
    exec(
        f"def {n}(x, *args, **kwargs): "
        f"return x._{n}_(*args, **kwargs) if hasattr(x, '_{n}_') else np_.{n}(x, *args, **kwargs)"
    )


Dates: TypeAlias = _da.Dater | Iterable[_da.Dater] | _da.Ranger | EllipsisType | None
Columns: TypeAlias = int | Iterable[int] | slice
Data: TypeAlias = Number | Iterable[Number] | tuple | np_.ndarray


def _get_date_positions(dates, base, num_periods):
    pos = list(_da.date_index(dates, base))
    min_pos = min(pos)
    max_pos = max(pos)
    add_before = max(-min_pos, 0)
    add_after = max(max_pos - num_periods, 0)
    pos_adjusted = [p + add_before for p in pos]
    return pos_adjusted, add_before, add_after


def _trim_decorate(func):
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        return self._trim()
    return wrapper


class Series(
    _sg.HodrickPrescottMixin,
    _ud.DescriptionMixin,
    _sc.ConversionMixin,
    _sx.EggsThirteenMixin,
    _vi.SeriesViewMixin,
    _sp.PlotlyMixin,
):
    """
    """
    __slots__ = (
        "start_date", "data", "data_type",
        "_description", "_column_titles", "_user_data",
    )
    _numeric_format: str = "15g"
    _short_str_format: str = ">15"
    _date_str_format: str = ">12"
    _missing = np_.nan
    _missing_str: str = "·"
    _test_missing_period = staticmethod(lambda x: np_.all(np_.isnan(x)))

    def __init__(self, /, *args, **kwargs):
        num_columns = kwargs.get("num_columns", 1)
        self.data_type = kwargs.get("data_type", np_.float64)
        self._description = kwargs.get("description", "")
        self.reset(num_columns)

    def reset(self, /, num_columns=None, data_type=None, description=None, ):
        num_columns = num_columns if num_columns else self.num_columns
        data_type = data_type if data_type else self.data_type
        description = description if description else self._description
        self.start_date = None
        self.data = np_.full((0, num_columns), self._missing, dtype=data_type)
        self._description = str(description)
        self._column_titles = [""] * num_columns
        self._user_data = {}
        return self

    def _create_periods_of_missing_values(self, num_rows=0):
        return np_.full((num_rows, self.shape[1]), self._missing, dtype=self.data_type)

    @property
    def shape(self):
        return self.data.shape

    @property
    def num_columns(self):
        return self.data.shape[1]

    @property
    def range(self):
        return _da.Ranger(self.start_date, self.end_date) if self.start_date else []

    @property
    def end_date(self):
        return self.start_date + self.data.shape[0] - 1 if self.start_date else None

    @property
    def frequency(self):
        return self.start_date.frequency if self.start_date is not None else _da.Frequency.UNKNOWN

    @classmethod
    def from_dates_and_data(
        cls,
        dates: Iterable[_da.Dater],
        data: np_.ndarray,
        /,
    ) -> Self:
        num_columns = data.shape[1] if hasattr(data, "shape") else 1
        self = cls(num_columns=num_columns)
        self.set_data(dates, data)
        return self

    from_dates_and_values = from_dates_and_data

    @classmethod
    def from_start_date_and_data(
        cls,
        start_date: _da.Dater,
        data: np_.ndarray,
        /,
        **kwargs,
    ) -> Self:
        num_columns = data.shape[1] if hasattr(data, "shape") else 1
        self = cls(num_columns=num_columns, **kwargs)
        self.start_date = start_date
        self.data = data
        return self._trim()

    @classmethod
    def from_func(
        cls,
        dates: Iterable[_da.Dater],
        func: Callable,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        self = cls(**kwargs)
        dates = [ t for t in dates ]
        data = [ [ func() for j in range(self.num_columns) ] for i in range(len(dates)) ]
        data = np_.array(data, dtype=self.data_type)
        self.set_data(dates, data)
        return self

    @_trim_decorate
    def set_data(
        self,
        dates: Dates,
        data: Data | Series,
        columns: Columns = None,
        /,
    ) -> Self:
        dates = self._resolve_dates(dates)
        columns = self._resolve_columns(columns)
        if not self.start_date:
            self.start_date = next(iter(dates))
            self.data = self._create_periods_of_missing_values(1)
        pos, add_before, add_after = _get_date_positions(dates, self.start_date, self.shape[0]-1)
        self.data = self._create_expanded_data(add_before, add_after)
        if add_before:
            self.start_date -= add_before
        if isinstance(data, Series):
            data = data.get_data(dates)
        if isinstance(data, np_.ndarray):
            self.data[pos, :] = data
            return
        if not isinstance(data, tuple):
            data = it_.repeat(data)
        for c, d in zip(columns, data):
            self.data[pos, c] = d
        return self

    def _get_data_and_recreate(
        self,
        *args,
    ) -> np_.ndarray:
        """
        """
        data, dates = self.get_data_and_resolved_dates(*args)
        num_columns = data.shape[1]
        new = Series(num_columns=num_columns, data_type=self.data_type)
        new.set_data(dates, data)
        return new

    def get_data_and_resolved_dates(
        self,
        dates: Dates,
        columns: Columns = None,
        /,
    ) -> np_.ndarray:
        """
        """
        dates = [ t for t in self._resolve_dates(dates) ]
        columns = self._resolve_columns(columns)
        base_date = self.start_date
        if not dates or not base_date:
            num_dates = len(set(dates))
            return self._create_periods_of_missing_values(num_dates)[:, columns], dates
        pos, add_before, add_after = _get_date_positions(dates, self.start_date, self.shape[0]-1)
        data = self._create_expanded_data(add_before, add_after)
        if not isinstance(pos, Iterable):
            pos = (pos, )
        return data[np_.ix_(pos, columns)], dates

    def get_data(
        self,
        *args,
    ) -> np_.ndarray:
        return self.get_data_and_resolved_dates(*args)[0]

    def get_data_column(
        self,
        dates: Dates,
        column: Number | None = None,
        /,
    ) -> np_.ndarray:
        """
        """
        column = column if column and column<self.data.shape[1] else 0
        return self.get_data(dates, column)

    def extract_columns(
        self,
        columns,
        /,
    ) -> None:
        if not isinstance(columns, Iterable):
            columns = (columns, )
        columns = [ c for c in columns ]
        self.data = self.data[:, columns]
        self.column_titles = [ self.column_titles[c] for c in columns ]

    def set_start_date(
        self,
        new_start_date: _da.Dater,
        /,
    ) -> Self:
        self.start_date = new_start_date
        return self

    def _resolve_dates(self, dates):
        if dates is None:
            return []
        if dates is Ellipsis:
            dates = _da.Ranger(None, None)
        if isinstance(dates, _da.ResolvableProtocol) and dates.needs_resolve:
            dates = dates.resolve(self)
        return dates

    def _resolve_columns(self, columns):
        if columns is None or columns is Ellipsis:
            columns = slice(None)
        if isinstance(columns, slice):
            columns = range(*columns.indices(self.num_columns))
        if not isinstance(columns, Iterable):
            columns = (columns, )
        return columns

    def __call__(self, *args):
        """
        Get data self[dates] or self[dates, columns]
        """
        return self.get_data(*args)

    def _shift(self, shift_by) -> Self:
        """
        Shift (lag, lead) start date
        """
        self.start_date = (
            self.start_date - shift_by 
            if self.start_date else self.start_date
        )

    def __getitem__(self, index):
        """
        Create a new time series based on date retrieved by self[dates] or self[dates, columns]
        """
        if isinstance(index, int):
            new = co_.deepcopy(self)
            new._shift(index)
            return new
        if not isinstance(index, tuple):
            index = (index, None, )
        return self._get_data_and_recreate(*index)

    def __setitem__(self, index, data):
        """
        Set data self[dates] = ... or self[dates, columns] = ...
        """
        if not isinstance(index, tuple):
            index = (index, None, )
        return self.set_data(index[0], data, index[1])

    def hstack(self, *args):
        if not args:
            return co_.deepcopy(self)
        encompassing_range = _da.get_encompassing_range(self, *args)
        new_data = self.get_data(encompassing_range)
        add_data = (
            x.get_data(encompassing_range) if hasattr(x, "get_data") else _create_data_from_number(x, encompassing_range, self.data_type)
            for x in args
        )
        new_data = np_.hstack((new_data, *add_data))
        new = Series(num_columns=new_data.shape[1]);
        new.set_data(encompassing_range, new_data)
        return new

    def clip(
        self,
        new_start_date: _da.Dater | None,
        new_end_date: _da.Dater | None,
    ) -> None:
        if new_start_date is None or new_start_date < self.start_date:
            new_start_date = self.start_date
        if new_end_date is None or new_end_date > self.end_date:
            new_end_date = self.end_date
        if new_start_date == self.start_date and new_end_date == self.end_date:
            return
        self.start_date = new_start_date
        self.data = self.get_data(_da.Ranger(new_start_date, new_end_date))

    def overlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        self._trim()
        other._trim()
        self.set_data(other.range, other.data, )
        return self

    def overlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        return self._LAY_METHOD_RESOLUTION[method]["overlay"](self, other, )

    def underlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        """
        """
        self._trim()
        other._trim()
        self_range = self.range
        self_data = self.data
        #
        self.start_date = other.start_date
        enforce_columns = np_.zeros((1, self.data.shape[1]), dtype=self.data_type)
        self.data = enforce_columns + other.data
        self.set_data(self_range, self_data, )
        return self

    def underlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        return self._LAY_METHOD_RESOLUTION[method]["underlay"](self, other, )

    _LAY_METHOD_RESOLUTION = {
        "by_range": {"overlay": overlay_by_range, "underlay": underlay_by_range},
    }

    def __or__(self, other):
        return self.hstack(other)

    def __lshift__(self, other):
        return co_.deepcopy(self).overlay_by_range(other, )

    def __rshift__(self, other):
        return co_.deepcopy(other).overlay_by_range(self, )

    def _trim(self):
        if self.data.size==0:
            return self.reset()
        num_leading = _get_num_leading_missing_rows(self.data, self._test_missing_period)
        if num_leading == self.data.shape[0]:
            return self.reset()
        num_trailing = _get_num_leading_missing_rows(self.data[::-1], self._test_missing_period)
        if not num_leading and not num_trailing:
            return self
        slice_from = num_leading if num_leading else None
        slice_to = -num_trailing if num_trailing else None
        self.data = self.data[slice(slice_from, slice_to), ...]
        if slice_from:
            self.start_date += int(slice_from)
        return self

    def _create_expanded_data(self, add_before, add_after):
        return np_.pad(
            self.data, ((add_before, add_after), (0, 0)),
            mode="constant", constant_values=self._missing
        )

    def _check_data_shape(self, data, /, ):
        if data.shape[1] != self.data.shape[1]:
            raise Exception("Time series data being assigned must preserve the number of columns")

    def __neg__(self):
        """
        -self
        """
        output = co_.deepcopy(self)
        output.data = -output.data
        return output

    def __pos__(self):
        """
        +self
        """
        return co_.deepcopy(self.copy)

    def __add__(self, other) -> Self|np_.ndarray:
        """
        self + other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x + y)

    def __radd__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__radd__(other))

    def __mul__(self, other) -> Self|np_.ndarray:
        """
        self + other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x * y)

    def __rmul__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__rmul__(other))

    def __sub__(self, other):
        """
        self - other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x - y)

    def __rsub__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rsub__(other))

    def __pow__(self, other):
        """
        self ** other
        """
        # FIXME: 1 ** numpy.nan -> 1 !!!!!
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x ** y)

    def __rpow__(self, other):
        """
        other ** self
        """
        return self.apply(lambda data: data.__rpow__(other))

    def __truediv__(self, other):
        """
        self - other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x / y)

    def __rtruediv__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rtruediv__(other))

    def __floordiv__(self, other):
        """
        self // other
        """
        return self._binop(other, lambda x, y: x // y)

    def __rfloordiv__(self, other):
        """
        other // self
        """
        return self.apply(lambda data: data.__rfloordiv__(other))

    def __mod__(self, other):
        """
        self % other
        """
        return self._binop(other, lambda x, y: x % y)

    def __rmod__(self, other, /, ):
        """
        other % self
        """
        return self.apply(lambda data: data.__rmod__(other))

    def apply(self, func, /, *args, **kwargs):
        new_data = func(self.data, *args, **kwargs)
        axis = kwargs.get("axis")
        if new_data.shape==self.data.shape:
            new = co_.deepcopy(self)
            new.data = new_data
            return new._trim()
        elif (axis is None or axis==1) and new_data.shape==(self.data.shape[0],):
            new = co_.deepcopy(self)
            new.data = new_data.reshape(self.data.shape[0], 1)
            return new._trim()
        elif axis is None and new_data.shape==():
            return new_data
        elif axis==0 and new_data.shape==(self.data.shape[1],):
            return new_data
        else:
            raise Exception("Function applied on a time series resulted in a data array with an unexpected shape")

    def _binop(self, other, func, /, ):
        if not isinstance(other, type(self)):
            unop_func = lambda data: func(data, other)
            return apply(self, unop_func)
        encompassing_range = _da.get_encompassing_range(self, other)
        self_data = self.get_data(encompassing_range)
        other_data = other.get_data(encompassing_range)
        new = Series()
        new.start_date = encompassing_range[0]
        new.data = func(self_data, other_data)
        new._trim()
        return new

    def _replace_data(self, new_data, /, ):
        self.data = new_data
        return self._trim()

    def copy(self, /, ) -> Self:
        return co_.deepcopy(self, )

    for n in ["log", "exp", "sqrt", "maximum", "minimum"]:
        exec(f"def {n}(self, *args, **kwargs, ): return self._replace_data(np_.{n}(self.data, *args, **kwargs, ), )")
        exec(f"def _{n}_(self, *args, **kwargs, ): return self.apply(np_.{n}, *args, **kwargs, )")

    for n in ["cumsum"]:
        exec(f"def {n}(self): return self._replace_data(np_.{n}(self.data, axis=0, ), )")

    for n in ["maxi", "mini", "mean", "median"]:
        exec(f"def _{n}_(self, *args, **kwargs): return self.apply(np_.{n}, *args, **kwargs, )")


def _get_num_leading_missing_rows(data, test_missing_period, /, ):
    try:
        num = next(
            i for i, period_data in enumerate(data) 
            if not test_missing_period(period_data)
        )
    except StopIteration:
        num = data.shape[0]
    return num


# def apply(x, func: Callable, /, *args, **kwargs) -> object:
    # if isinstance(x, Series):
        # return x.apply(func, *args, **kwargs)
    # else:
        # return func(x, *args, **kwargs)

def hstack(first, *args) -> Self:
    return first.hstack(*args)


def _create_data_from_number(
    number: Number,
    range: _da.Ranger,
    data_type: type,
    /,
) -> np_.ndarray:
    return np_.full((len(range), 1), number, dtype=data_type)


def shift(x, by):
    new = co_.deepcopy(x)
    new._shift(by)
    return new


def diff(x, by=-1):
    return x - shift(x, by)


def difflog(x, by=-1):
    return log(x) - log(shift(x, by))


def pct(x, shift=-1):
    return 100*(x/shift(x, by) - 1)


def roc(x, shift=-1):
    return x/shift(x, by)





"""
Dates, date ranges, and date frequencies
"""


#[
from __future__ import annotations
import re as _re
import datetime as _dt
import enum as _en
import functools as _ft
from typing import (Union, Self, Any, Protocol, TypeAlias, runtime_checkable, )
from collections.abc import (Iterable, Callable, )
from numbers import (Number, )

from ..user import copy as _cp
#]


__all__ = [
    "Frequency",
    "yy", "hh", "qq", "mm", "dd", "ii",
    "Ranger", "start", "end",
    "dater_from_sdmx_string",
    "daters_from_sdmx_strings",
    "dater_from_iso_string",
    "daters_from_iso_strings",
]


class Frequency(_en.IntEnum):
    """
    Enumeration of date frequencies
    """
    #[
    INTEGER = 0
    YEARLY = 1
    HALFYEARLY = 2
    QUARTERLY = 4
    MONTHLY = 12
    WEEKLY = 52
    DAILY = 365
    UNKNOWN = -1

    @classmethod
    def from_letter(cls, letter: str, /, ) -> Frequency:
        """
        """
        return next(
            x for x in cls
            if x.name.startswith(letter.upper()) and x is not cls.UNKNOWN
        )

    @property
    def letter(self, /, ) -> str:
        return self.name[0] if self is not self.UNKNOWN else "?"

    @property
    def plotly_format(self, /, ) -> str:
        return PLOTLY_FORMATS[self]

    @property
    def is_regular(self, /, ) -> bool:
        return self in [self.YEARLY, self.HALFYEARLY, self.QUARTERLY, self.MONTHLY]

    def __str__(self, /, ) -> str:
        return self.name
    #]


PLOTLY_FORMATS = {
    Frequency.INTEGER: None,
    Frequency.YEARLY: "%Y",
    Frequency.HALFYEARLY: "%Y-%m",
    Frequency.QUARTERLY: "%Y-Q%q",
    Frequency.MONTHLY: "%Y-%m",
    Frequency.WEEKLY: "%Y-%W",
    Frequency.DAILY: "%Y-%m-%d",
}


BASE_YEAR = 2020


@runtime_checkable
class ResolutionContextProtocol(Protocol, ):
    """
    Context protocol for contextual date resolution
    """
    start_date = ...
    end_date = ...


@runtime_checkable
class ResolvableProtocol(Protocol, ):
    """
    Contextual date protocol
    """
    needs_resolve = ...
    def resolve(self, context: ResolutionContextProtocol) -> Any: ...


def _check_daters(first, second) -> None:
    if type(first) is not type(second):
        raise Exception("Dates must be the same date frequency")


def _check_daters_decorate(func: Callable, ) -> Callable:
    def wrapper(*args, **kwargs):
        _check_daters(args[0], args[1], )
        return func(*args, **kwargs, )
    return wrapper


def _check_offset(offset, ) -> None:
    if not isinstance(offset, int, ):
        raise Exception("Date offset must be an integer")


def _check_offset_decorator(func: Callable) -> Callable:
    def wrapper(*args, **kwargs):
        _check_offset(args[1])
        return func(*args, **kwargs)
    return wrapper


def _remove_blanks(func: Callable,) -> Callable:
    def wrapper(*args, **kwargs, ):
        return func(*args, **kwargs, ).replace(" ", "", )
    return wrapper


class RangeableMixin:
    #[
    def __rshift__(self, end_date: Self) -> Ranger:
        """
        """
        return Ranger(self, end_date, 1)

    def __lshift__(self, start_date: Self) -> Ranger:
        """
        """
        return Ranger(start_date, self, -1) 
    #]


class IsoMixin:
    #[
    def to_iso_string(
        self,
        /,
        position: Literal["start"] | Literal["center"] | Literal["end"] = "start",
    ) -> str:
        year, month, day = self.to_ymd(position=position, )
        return f"{year:04g}-{month:02g}-{day:02g}"

    @classmethod
    def from_iso_string(cls: type, iso_string: str, ) -> Self:
        year, month, day = iso_string.split("-", )
        return cls.from_ymd(int(year), int(month), int(day), )

    to_plotly_date = _ft.partialmethod(to_iso_string, position="center")
    #]


class Dater(
    RangeableMixin,
    _cp.CopyMixin,
):
    """
    """
    #[
    frequency = None
    needs_resolve = False

    def __init__(self, serial=0) -> None:
        self.serial = int(serial)

    def get_distance_from_origin(self) -> int:
        return self.serial - self.origin

    def resolve(self, context: ResolutionContextProtocol) -> Self:
        return self

    def __bool__(self) -> bool:
        return not self.needs_resolve

    def __len__(self):
        return 1

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.to_sdmx_string()

    _databank_repr = __str__

    def __format__(self, *args) -> str:
        str_format = args[0] if args else ""
        return ("{date_str:"+str_format+"}").format(date_str=self.__str__())

    def __iter__(self) -> Iterable[Self]:
        yield self

    def _get_hash_key(self, /, ) -> tuple[int, int]:
        return (int(self.serial), int(self.frequency), )

    def __hash__(self, /, ) -> int:
        return hash(self._get_hash_key(), )

    @_check_offset_decorator
    def __add__(self, other: int) -> Self:
        return type(self)(self.serial + int(other))

    __radd__ = __add__

    def __sub__(self, other: Union[Self, int]) -> Union[Self, int]:
        if isinstance(other, Dater):
            return self._sub_dater(other)
        else:
            return self.__add__(-int(other))

    @_check_daters_decorate
    def _sub_dater(self, other: Self) -> int:
        return self.serial - other.serial

    def __index__(self):
        return self.serial

    @_check_daters_decorate
    def __eq__(self, other) -> bool:
        return self.serial == other.serial

    @_check_daters_decorate
    def __ne__(self, other) -> bool:
        return self.serial != other.serial

    @_check_daters_decorate
    def __lt__(self, other) -> bool:
        return self.serial < other.serial

    @_check_daters_decorate
    def __le__(self, other) -> bool:
        return self.serial <= other.serial

    @_check_daters_decorate
    def __gt__(self, other) -> bool:
        return self.serial > other.serial

    @_check_daters_decorate
    def __ge__(self, other) -> bool:
        return self.serial >= other.serial
    #]


class IntegerDater(Dater, ):
    #[
    frequency = Frequency.INTEGER
    needs_resolve = False
    origin = 0

    @classmethod
    def from_sdmx_string(cls, sdmx_string: str) -> IntegerDater:
        sdmx_string = sdmx_string.strip().removeprefix("(", "").removesuffix(")")
        return cls(int(sdmx_string))

    def to_sdmx_string(self) -> str:
        return f"({self.serial})"

    def to_plotly_date(self) -> str:
        return str(self.serial)

    def __repr__(self) -> str:
        return f"ii({self.serial})"

    def to_plotly_date(self):
        return self.serial
    #]


class DailyDater(Dater, IsoMixin):
    #[
    frequency: Frequency = Frequency.DAILY
    needs_resolve = False
    origin = _dt.date(BASE_YEAR, 1, 1).toordinal()

    @classmethod
    def from_ymd(cls: type, year: int, month: int=1, day: int=1) -> Self:
        serial = _dt.date(year, month, day).toordinal()
        return cls(serial)

    @classmethod
    def from_year_period(cls, year: int, period: int) -> Self:
        boy_serial = _dt.date(year, 1, 1).toordinal()
        serial = boy_serial + int(period) - 1
        return cls(serial)

    @classmethod
    def from_sdmx_string(cls, sdmx_string: str) -> DailyDater:
        year, month, day, *_ = sdmx_string.split("-")
        return cls.from_ymd(int(year), int(month), int(day))

    def to_sdmx_string(self, /, **kwargs) -> str:
        year, month, day = self.to_ymd()
        return f"{year:04g}-{month:02g}-{day:02g}"

    def to_year_period(self) -> tuple[int, int]:
        boy_serial = _dt.date(_dt.date.fromordinal(self.serial).year, 1, 1)
        per = self.serial - boy_serial + 1
        year = _dt.date.fromordinal(self.serial).year
        return year, per

    def to_ymd(self, /, **kwargs) -> tuple[int, int, int]:
        py_date = _dt.date.fromordinal(self.serial)
        return py_date.year, py_date.month, py_date.day

    def get_year(self, /, ) -> int:
        return _dt.date.fromordinal(self.serial).year

    @_remove_blanks
    def __repr__(self) -> str:
        return f"dd{self.to_ymd()}"

    def to_start_of_year(self, ) -> Self:
        year = self.get_year()
        serial = _dt.date(year, 1, 1).toordinal()
        return type(self)(serial)

    def to_end_of_year(self, ) -> Self:
        year = self.get_year()
        serial = _dt.date(year, 12, 31).toordinal()
        return type(self)(serial)

    def to_end_of_previous_year(self, ) -> Self:
        year = self.get_year()
        serial = _dt.date(year-1, 12, 31).toordinal()
        return type(self)(serial)

    def to_daily(self, /, **kwargs, ) -> Self:
        return self
    #]


def _serial_from_ypf(year: int, per: int, freq: int) -> int:
    return int(year)*int(freq) + int(per) - 1


class RegularDaterMixin(IsoMixin, ):
    #[
    @classmethod
    def from_year_period(
            cls: type,
            year: int,
            per: int | str = 1,
        ) -> Self:
        per = per if per != "end" else cls.frequency.value
        new_serial = _serial_from_ypf(year, per, cls.frequency.value)
        return cls(new_serial)

    @classmethod
    def from_ymd(cls, year: int, month: int=1, day: int=1, ) -> YearlyDater:
        return cls.from_year_period(year, cls.month_to_period(month, ), )

    def to_year_period(self) -> tuple[int, int]:
        return self.serial//self.frequency.value, self.serial%self.frequency.value+1

    def get_year(self) -> int:
        return self.to_year_period()[0]

    def to_ymd(
        self, 
        /,
        position: Literal["start"] | Literal["center"] | Literal["end"] = "center",
    ) -> tuple[int, int, int]:
        year, per = self.to_year_period()
        return year, *self.month_day_resolution[position][per]

    def __str__(self) -> str:
        year, per = self.to_year_period()
        letter = self.frequency.letter
        return self.frequency.sdmx_format.format(year=year, per=per, letter=letter)

    def to_start_of_year(self, ) -> Self:
        year, *_ = self.to_year_period()
        return self.from_year_period(year, 1)

    def to_end_of_year(self, ) -> Self:
        year, *_ = self.to_year_period()
        return self.from_year_period(year, self.frequency.value)

    def to_end_of_previous_year(self, ) -> Self:
        year, *_ = self.to_year_period()
        return self.from_year_period(year-1, self.frequency.value)

    def to_daily(
        self,
        /,
        position: Literal["start"] | Literal["center"] | Literal["end"] = "center"
    ) -> DailyDater:
        return DailyDater.from_ymd(*self.to_ymd(position=position))
    #]


class YearlyDater(Dater, RegularDaterMixin, ): 
    #[
    frequency: Frequency = Frequency.YEARLY
    needs_resolve: bool = False
    origin = _serial_from_ypf(BASE_YEAR, 1, Frequency.YEARLY)
    month_day_resolution = {
        "start": {1: (1, 1)},
        "center": {1: (6, 30)},
        "end": {1: (12, 31)},
    }

    @classmethod
    def from_sdmx_string(cls, sdmx_string: str) -> YearlyDater:
        return cls(int(sdmx_string))

    def to_sdmx_string(self) -> str:
        return f"{self.get_year():04g}"

    @_remove_blanks
    def __repr__(self) -> str: return f"yy({self.get_year()})"

    @staticmethod
    def month_to_period(month: int, ) -> int:
        return 1
    #]


class HalfyearlyDater(Dater, RegularDaterMixin, ):
    #[
    frequency: Frequency = Frequency.HALFYEARLY
    needs_resolve: bool = False
    origin = _serial_from_ypf(BASE_YEAR, 1, Frequency.HALFYEARLY)
    month_day_resolution = {
        "start": {1: (1, 1), 2: (7, 1)},
        "center": {1: (3, 15), 2: (9, 15)},
        "end": {1: (6, 30), 2: (12, 31)}
    }

    @classmethod
    def from_sdmx_string(cls, sdmx_string: str) -> HalfyearlyDater:
        year, halfyear = sdmx_string.split("-H")
        return cls.from_year_period(int(year), int(halfyear))

    def to_sdmx_string(self) -> str:
        year, per = self.to_year_period()
        return f"{year:04g}-{self.frequency.letter}{per:1g}"

    @_remove_blanks
    def __repr__(self) -> str: return f"hh{self.to_year_period()}"

    def to_ymd(
        self, 
        /,
        position: Literal["start"] | Literal["center"] | Literal["end"] = "center",
    ) -> tuple[int, int, int]:
        year, per = self.to_year_period()
        return (
            year,
            *{"start": (1, 1), "center": (6, 3), "end": (12, 31)}[position],
        )

    def get_month(
        self,
        /,
        position: Literal["start"] | Literal["center"] | Literal["end"] = "center",
    ) -> int:
        _, per = self.to_year_period()
        return month_resolution[position][per]

    @staticmethod
    def month_to_period(month: int, ) -> int:
        return 1+((month-1)//6)
    #]


class QuarterlyDater(Dater, RegularDaterMixin, ):
    frequency: Frequency = Frequency.QUARTERLY
    needs_resolve: bool = False
    origin = _serial_from_ypf(BASE_YEAR, 1, Frequency.QUARTERLY)
    month_day_resolution = {
        "start": {1: (1, 1), 2: (4, 1), 3: (7, 1), 4: (10, 1)},
        "center": {1: (2, 15), 2: (5, 15), 3: (8, 15), 4: (11, 15)},
        "end": {1: (3, 30), 2: (6, 31), 3: (9, 31), 4: (12, 31)},
    }

    @classmethod
    def from_sdmx_string(cls, sdmx_string: str) -> QuarterlyDater:
        year, quarter = sdmx_string.split("-Q")
        return cls.from_year_period(int(year), int(quarter))

    def to_sdmx_string(self) -> str:
        year, per = self.to_year_period()
        return f"{year:04g}-{self.frequency.letter}{per:1g}"

    @_remove_blanks
    def __repr__(self) -> str: return f"qq{self.to_year_period()}"

    @staticmethod
    def month_to_period(month: int, ) -> int:
        return 1+((month-1)//3)


class MonthlyDater(Dater, RegularDaterMixin, ):
    #[
    frequency: Frequency = Frequency.MONTHLY
    needs_resolve: bool = False
    origin = _serial_from_ypf(BASE_YEAR, 1, Frequency.MONTHLY)
    month_day_resolution = {
        "start": {1: (1, 1), 2: (2, 1), 3: (2, 1), 4: (4, 1), 5: (5, 1), 6: (6, 1), 7: (7, 1), 8: (8, 1), 9: (9, 1), 10: (10, 1), 11: (11, 1), 12: (12, 1)},
        "center": {1: (1, 15), 2: (2, 15), 3: (2, 15), 4: (4, 15), 5: (5, 15), 6: (6, 15), 7: (7, 15), 8: (8, 15), 9: (9, 15), 10: (10, 15), 11: (11, 15), 12: (12, 15)},
        "end": {1: (1, 31), 2: (2, 28), 3: (2, 31), 4: (4, 30), 5: (5, 31), 6: (6, 30), 7: (7, 31), 8: (8, 31), 9: (9, 30), 10: (10, 31), 11: (11, 30), 12: (12, 31)},
    }

    @classmethod
    def from_sdmx_string(cls, sdmx_string: str) -> MonthlyDater:
        year, month = sdmx_string.split("-")
        return cls.from_year_period(int(year), int(month))

    def to_sdmx_string(self) -> str:
        year, per = self.to_year_period()
        return f"{year:04g}-{per:02g}"

    @_remove_blanks
    def __repr__(self) -> str: return f"mm{self.to_year_period()}"

    @staticmethod
    def month_to_period(month: int, ) -> int:
        return month
    #]


yy = YearlyDater.from_year_period

hh = HalfyearlyDater.from_year_period

qq = QuarterlyDater.from_year_period

mm = MonthlyDater.from_year_period

ii = IntegerDater




def dd(year: int, month: int | ellipsis, day: int) -> DailyDater:
    if month is Ellipsis:
        return DailyDater.from_year_period(year, day)
    else:
        return DailyDater.from_ymd(year, month, day)


class Ranger():
    #[
    def __init__(
        self, 
        start_date: Dater|None =None,
        end_date: Dater|None =None,
        step: int=1,
        /
    ) -> None:
        """
        Date range constructor
        """
        start_date = resolve_dater_or_integer(start_date)
        end_date = resolve_dater_or_integer(end_date)
        self._start_date = start_date if start_date is not None else start
        self._end_date = end_date if end_date is not None else end
        self._step = step
        self.needs_resolve = self._start_date.needs_resolve or self._end_date.needs_resolve
        if not self.needs_resolve:
            _check_daters(start_date, end_date)

    @property
    def start_date(self):
        return self._start_date

    @property
    def end_date(self):
        return self._end_date

    @property
    def step(self):
        return self._step

    @property
    def _class(self):
        return type(self._start_date) if not self.needs_resolve else None

    @property
    def _serials(self) -> range|None:
        return range(self._start_date.serial, self._end_date.serial+_sign(self._step), self._step) if not self.needs_resolve else None

    # @property
    # def needs_resolve(self) -> bool:
        # return bool(self._start_date and self._end_date)

    @property
    def frequency(self) -> Frequency:
        return self._class.frequency

    def to_plotly_dates(self) -> Iterable[str]:
        return [t.to_plotly_date() for t in self]

    def __len__(self) -> int|None:
        return len(self._serials) if not self.needs_resolve else None

    def __str__(self) -> str:
        step_str = f", {self._step}" if self._step!=1 else ""
        start_date_str = self._start_date.__str__()
        end_date_str = self._end_date.__str__()
        return f"Ranger({start_date_str}, {end_date_str}{step_str})"

    def __repr__(self) -> str:
        step_rep = f", {self._step}" if self._step!=1 else ""
        start_date_rep = self._start_date.__repr__()
        end_date_rep = self._end_date.__repr__()
        return f"Ranger({start_date_rep}, {end_date_rep}{step_rep})"

    def __add__(self, offset: int) -> range:
        return Ranger(self._start_date+offset, self._end_date+offset, self._step)

    __radd__ = __add__

    def __sub__(self, offset: Union[Dater, int]) -> Union[range, Self]:
        if isinstance(offset, Dater):
            return range(self._start_date-offset, self._end_date-offset, self._step) if not self.needs_resolve else None
        else:
            return Ranger(self._start_date-offset, self._end_date-offset, self._step)

    def __rsub__(self, ather: Dater) -> range|None:
        if isinstance(other, Dater):
            return range(other-self._start_date, other-self._end_date, -self._step) if not self.needs_resolve else None
        else:
            return None

    def __iter__(self) -> Iterable:
        return (self._class(x) for x in self._serials) if not self.needs_resolve else None

    def __getitem__(self, i: int) -> Dater|None:
        return self._class(self._serials[i]) if not self.needs_resolve else None

    def resolve(self, context: ResolutionContextProtocol) -> Self:
        resolved_start_date = self._start_date if self._start_date else self._start_date.resolve(context)
        resolved_end_date = self._end_date if self._end_date else self._end_date.resolve(context)
        return Ranger(resolved_start_date, resolved_end_date, self._step)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def __eq__(self, other: Self) -> bool:
        return self._start_date==other._start_date and self._end_date==other._end_date and self._step==other._step

    def __bool__(self) -> bool:
        return not self.needs_resolve
    #]


def _sign(x: Number) -> int:
    return 1 if x>0 else (0 if x==0 else -1)


def date_index(dates: Iterable[Dater], base: Dater) -> Iterable[int]:
    return (x-base for x in dates)



class ContextualDater(Dater, RangeableMixin):
    """
    Dates with context dependent resolution
    """
    #[
    needs_resolve = True

    def __init__(self, resolve_from: str, offset: int=0) -> None:
        self._resolve_from = resolve_from
        self._offset = offset

    def __add__(self, offset: int) -> None:
        return type(self)(self._resolve_from, self._offset+offset)

    def __sub__(self, offset: int) -> None:
        return type(self)(self._resolve_from, self._offset-offset)

    def __str__(self) -> str:
        return "<>." + self._resolve_from + (f"{self._offset:+g}" if self._offset else "")

    def __repr__(self) -> str:
        return self.__str__()

    def __bool__(self) -> bool:
        return False

    def resolve(self, context: ResolutionContextProtocol) -> Dater:
        return getattr(context, self._resolve_from) + self._offset
    #]


start = ContextualDater("start_date")
end = ContextualDater("end_date")


def resolve_dater_or_integer(input_date: Any) -> Dater:
    """
    Convert non-dater to integer dater
    """
    if isinstance(input_date, Number):
        input_date = IntegerDater(int(input_date))
    return input_date


DATER_CLASS_FROM_FREQUENCY_RESOLUTION = {
    Frequency.INTEGER: IntegerDater,
    Frequency.YEARLY: YearlyDater,
    Frequency.HALFYEARLY: HalfyearlyDater,
    Frequency.QUARTERLY: QuarterlyDater,
    Frequency.MONTHLY: MonthlyDater,
    Frequency.DAILY: DailyDater,
}


def dater_from_sdmx_string(freq: Frequency, sdmx_string: str, ) -> Dater:
    """
    """
    return DATER_CLASS_FROM_FREQUENCY_RESOLUTION[freq].from_sdmx_string(sdmx_string)


def daters_from_sdmx_strings(freq: Frequency, sdmx_strings: Iterable[str], ) -> Iterable[Dater]:
    """
    """
    return (DATER_CLASS_FROM_FREQUENCY_RESOLUTION[freq].from_sdmx_string(x) for x in sdmx_strings)


def dater_from_iso_string(freq: Frequency, iso_string: str, ) -> Dater:
    """
    """
    return DATER_CLASS_FROM_FREQUENCY_RESOLUTION[freq].from_iso_string(iso_string)


def daters_from_iso_strings(freq: Frequency, iso_strings: Iterable[str], ) -> Iterable[Dater]:
    """
    """
    return (DATER_CLASS_FROM_FREQUENCY_RESOLUTION[freq].from_iso_string(x) for x in iso_strings)


def dater_from_ymd(freq: Frequency, *args, ) -> Dater:
    """
    """
    return DATER_CLASS_FROM_FREQUENCY_RESOLUTION[freq].from_ymd(*args, )


_FREQUENCY_FROM_STRING_RESOLUTION = {
    "I": Frequency.INTEGER,
    "Y": Frequency.YEARLY,
    "H": Frequency.HALFYEARLY,
    "Q": Frequency.QUARTERLY,
    "M": Frequency.MONTHLY,
    "W": Frequency.WEEKLY,
    "D": Frequency.DAILY,
}


def frequency_from_string(text: str) -> Frequency:
    """
    """
    first_letter_match = _re.search("[A-Z]", text.upper() + "X")
    return _FREQUENCY_FROM_STRING_RESOLUTION.get(first_letter_match.group(0), Frequency.UNKNOWN)


def get_encompassing_range(*args: ResolutionContextProtocol, ) -> Ranger:
    start_dates = [x.start_date for x in args if hasattr(x, "start_date") and x.start_date]
    end_dates = [x.end_date for x in args if hasattr(x, "end_date") and x.end_date]
    start_date = min(start_dates) if start_dates else None
    end_date = max(end_dates) if end_dates else None
    return Ranger(start_date, end_date)




"""
# First-order solution matrices
"""


#[
import enum as en_
import numpy as np_
import scipy as sp_
from typing import (Self, Callable, )
from numbers import (Number, )
import dataclasses as _dc

from ..fords import (systems as _fy, descriptors as _fd, )
from ..models import (flags as _mg, )
#]


class EigenValueKind(en_.Flag):
    STABLE = en_.auto()
    UNIT = en_.auto()
    UNSTABLE = en_.auto()


class SystemStabilityKind(en_.Flag):
    STABLE = en_.auto()
    MULTIPLE_STABLE = en_.auto()
    NO_STABLE = en_.auto()


@_dc.dataclass(slots=True, )
class Solution:
    """
    ## Square solution:
    T: Transition matrix
    R: Impact matrix of transition shocks
    K: Intercept in transition equation
    Z: Measurement matrix
    H: Impact matrix of measurement shocks
    D: Intercept in measurement equation

    ## Triangular solution:
    Ta: Transition matrix in triangular system
    Ra: Impact matrix of transition shocks in triangular system
    Ka: Intercept in transition equation in triangular system
    Za: Measurement matrix in triangular system

    ## Forward expansion of square solution:
    J: Power matrix
    Ru: Forward-looking impact matrix of transition shocks
    X: Impact matrix in square system
    Xa: Impact matrix in triangular system
    """
    #[
    T: np_.ndarray | None = None
    R: np_.ndarray | None = None
    K: np_.ndarray | None = None
    Z: np_.ndarray | None = None
    H: np_.ndarray | None = None
    D: np_.ndarray | None = None

    Ta: np_.ndarray | None = None
    Ra: np_.ndarray | None = None
    Ka: np_.ndarray | None = None
    Za: np_.ndarray | None = None
    Ua: np_.ndarray | None = None

    J: np_.ndarray | None = None
    Ru: np_.ndarray | None = None
    X: np_.ndarray | None = None
    Xa: np_.ndarray | None = None

    eigen_values: tuple[Number, ...] | None = None
    eigen_values_stability: tuple[EigenValueKind, ...] | None = None
    system_stability: SystemStabilityKind | None = None

    def __init__(
        self, 
        descriptor: _fd.Descriptor,
        system: _fy.System,
        model_flags: _mg.Flags,
        /,
        *,
        tolerance: float = 1e-12,
    ) -> Self:
        """
        """
        is_alpha_beta_stable_or_unit_root = lambda alpha, beta: abs(beta) < (1 + tolerance)*abs(alpha)
        is_stable_root = lambda root: abs(root) < (1 - tolerance)
        is_unit_root = lambda root: abs(root) >= (1 - tolerance) and abs(root) < (1 + tolerance)
        #
        # Detach unstable from (stable + unit) roots and solve out expectations
        qz, eigen_values, eigen_values_stability = _solve_ordqz(system, is_alpha_beta_stable_or_unit_root, is_stable_root, is_unit_root, )
        system_stability = _classify_system_stability(descriptor, eigen_values_stability, )
        triangular_solution_prelim = _solve_transition_equations(descriptor, system, qz, )
        #
        # Detach unit from stable roots and transform to square form
        triangular_solution = detach_stable_from_unit_roots(triangular_solution_prelim, is_unit_root, )
        square_solution = _square_from_triangular(triangular_solution, )
        self.Ua, self.Ta, self.Ra, self.Ka, self.Xa, self.J, self.Ru = triangular_solution
        self.T, self.R, self.K, self.X = square_solution
        #
        # Solve measurement equations
        self.Z, self.H, self.D, self.Za = _solve_measurement_equations(descriptor, system, self.Ua, )
        self.eigen_values, self.eigen_values_stability = eigen_values, eigen_values_stability
        self.system_stability = system_stability
        #
        return self

    def expand_square_solution(self, forward, /, ) -> list[np_.ndarray]:
        """
        Expand R matrices of square solution for t+1...t+forward
        """
        R, X, J, Ru = self.R, self.X, self.J, self.Ru
        if (R is None) or (X is None) or (J is None) or (Ru is None):
            return None
        Jk = np_.eye(J.shape[0])
        #
        # return [R(t+1), R(t+2), ..., R(t+forward)]
        #
        # R(t+k) = -X J**(k-1) Ru e(t+k)
        # k = 1, ..., forward or k-1 = 0, ..., forward-1
        #
        return [
            -X @ np_.linalg.matrix_power(J, k_minus_1) @ Ru 
            for k_minus_1 in range(0, forward)
        ]
    #]


def _left_div(A, B):
    """
    Solve A \ B
    """
    return np_.linalg.lstsq(A, B, rcond=None)[0]


def _right_div(B, A):
    """
    Solve B/A = (A'\B')'
    """
    return np_.linalg.lstsq(A.T, B.T, rcond=None)[0].T


def _square_from_triangular(
    triangular_solution: tuple[np_.ndarray, ...],
    /,
) -> tuple[np_.ndarray, ...]:
    """
    T <- Ua @ Ta/U; note that Ta/U == (U'\Ta')'
    R <- Ua @ Ra;
    X <- Xa @ Ra;
    K <- Ua @ Ka;
    xi(t) = ... -X J**(k-1) Ru e(t+k)
    """
    #[
    Ua, Ta, Ra, Ka, Xa, *_ = triangular_solution
    T = Ua @ _right_div(Ta, Ua) # Ua @ Ta / Ua
    R = Ua @ Ra
    K = Ua @ Ka
    X = Ua @ Xa
    return T, R, K, X
    #]


def detach_stable_from_unit_roots(
    transition_solution_prelim: tuple[np_.ndarray, ...],
    is_unit_root: Callable[[Number], bool],
    /,
) -> tuple[np_.ndarray, ...]:
    """
    """
    #[
    Ug, Tg, Rg, Kg, Xg, J, Ru = transition_solution_prelim
    num_xib = Tg.shape[0]
    Ta, u, check_num_unit_roots = sp_.linalg.schur(Tg, sort=is_unit_root, ) # Tg = u @ Ta @ u.T
    Ua = Ug @ u if Ug is not None else u
    Ra = u.T @ Rg
    Ka = u.T @ Kg
    Xa = u.T @ Xg if Xg is not None else None
    return Ua, Ta, Ra, Ka, Xa, J, Ru
    #]


def _solve_measurement_equations(descriptor, system, Ua, ) -> tuple[np_.ndarray, ...]:
    """
    """
    #[
    num_forwards = descriptor.get_num_forwards()
    G = system.G[:, num_forwards:]
    Z = _left_div(-system.F, G) # -F \ G
    H = _left_div(-system.F, system.J) # -F \ J
    D = _left_div(-system.F, system.H) # -F \ H
    Za = Z @ Ua
    return Z, H, D, Za
    #]


def _solve_transition_equations(descriptor, system, qz, ) -> tuple[np_.ndarray, ...]:
    """
    """
    #[
    num_backwards = descriptor.get_num_backwards()
    num_forwards = descriptor.get_num_forwards()
    num_stable = num_backwards
    S, T, Q, Z = qz
    A, B, C, D = system.A, system.B, system.C, system.D
    #
    S11 = S[:num_stable, :num_stable]
    S12 = S[:num_stable, num_stable:]
    S22 = S[num_stable:, num_stable:]
    #
    T11 = T[:num_stable, :num_stable]
    T12 = T[:num_stable, num_stable:]
    T22 = T[num_stable:, num_stable:]
    #
    Z21 = Z[num_forwards:, :num_stable]
    Z22 = Z[num_forwards:, num_stable:]
    #
    QC = Q @ C
    QC1 = QC[:num_stable, ...]
    QC2 = QC[num_stable:, ...]
    #
    QD = Q @ D
    QD1 = QD[:num_stable, ...]
    QD2 = QD[num_stable:, ...]
    #
    # Unstable block
    #
    G = _left_div(-Z21, Z22) # -Z21 \ Z22
    Ru = _left_div(-T22, QD2) # -T22 \ QD2
    Ku = _left_div(-(S22 + T22), QC2) # -(S22+T22) \ QC2
    #
    # Transform stable block==transform backward-looking variables:
    # gamma(t) = s(t) + G u(t+1)
    #
    Xg0 = _left_div(S11, T11 @ G + T12)
    Xg1 = G + _left_div(S11, S12)
    #
    Tg = _left_div(-S11, T11)
    Rg = -Xg0 @ Ru - _left_div(S11, QD1)
    Kg = -(Xg0 + Xg1) @ Ku - _left_div(S11, QC1)
    Ug = Z21 # xib = Ug @ gamma
    #
    # Forward expansion
    # gamma(t) = ... -Xg J**(k-1) Ru e(t+k)
    #
    J = _left_div(-T22, S22) # -T22 \ S22
    Xg = Xg1 + Xg0 @ J
    #
    return Ug, Tg, Rg, Kg, Xg, J, Ru
    #]


def _solve_ordqz(
    system,
    is_alpha_beta_stable_or_unit_root,
    is_stable_root,
    is_unit_root,
    /,
) -> tuple[tuple[np_.ndarray, ...], tuple[Number, ], tuple[EigenValueKind, ...]]:
    """
    """
    #[
    S, T, alpha, beta, Q, Z = sp_.linalg.ordqz(system.A, system.B, sort=is_alpha_beta_stable_or_unit_root, )
    Q = Q.T
    #
    inx_nonzero_alpha = alpha != 0
    eigen_values = np_.full(beta.shape, np_.inf, dtype=complex, )
    eigen_values[inx_nonzero_alpha] = -beta[inx_nonzero_alpha] / alpha[inx_nonzero_alpha]
    eigen_values = tuple(eigen_values)
    #
    eigen_values_stability = tuple(
        _classify_eig_value_stability(v, is_stable_root, is_unit_root, )
        for v in eigen_values
    )
    #
    return (S, T, Q, Z), eigen_values, eigen_values_stability
    #]


def _classify_eig_value_stability(eig_value, is_stable_root, is_unit_root, ) -> EigenValueKind:
    #[
    abs_eig_value = np_.abs(eig_value)
    if is_stable_root(abs_eig_value):
        return EigenValueKind.STABLE
    elif is_unit_root(abs_eig_value):
        return EigenValueKind.UNIT
    else:
        return EigenValueKind.UNSTABLE
    #]


def _classify_system_stability(descriptor, eigen_values_stability, ):
    #[
    num_unstable = sum(1 for s in eigen_values_stability if s==EigenValueKind.UNSTABLE)
    num_forwards = descriptor.get_num_forwards()
    if num_unstable == num_forwards:
        stability = SystemStabilityKind.STABLE
    elif num_unstable > num_forwards:
        stability = SystemStabilityKind.NO_STABLE
    else:
        stability = SystemStabilityKind.MULTIPLE_STABLE
    return stability
    #]





"""
# First-order unsolved system matrices

## Unsolved system

$$
A E[x_t] + B E[x_{t-1}] + C + D v_t = 0 \\
F y_t + G x_t + H + J w_t = 0
$$
"""


#[
from __future__ import annotations

import dataclasses as dc_
import numpy as np_ 

from . import (descriptors as fd_, )
#]


@dc_.dataclass
class System:
    """
    Unsolved system matrices
    """
    #
    # Transition equations
    #
    A: np_.ndarray | None = None
    B: np_.ndarray | None = None
    C: np_.ndarray | None = None
    D: np_.ndarray | None = None
    #
    # Measurement equations
    # F y + G xi + H + J e = 0
    #
    F: np_.ndarray | None = None
    G: np_.ndarray | None = None
    H: np_.ndarray | None = None
    J: np_.ndarray | None = None

    def __init__(
        self,
        descriptor: fd_.Descriptor,
        data_array: np_.ndarray,
        steady_array: np_.ndarray,
        /,
    ) -> None:
        """
        """
        # Differentiate and evaluate constant
        td, tc = descriptor.aldi_context.eval_to_arrays(
            data_array, steady_array,
        )

        smap = descriptor.system_map
        svec = descriptor.system_vectors

        self.A = np_.zeros(svec.shape_A_excl_dynid, dtype=float)
        self.A[smap.A.lhs] = td[smap.A.rhs]
        self.A = np_.vstack((self.A, smap.dynid_A))

        self.B = np_.zeros(svec.shape_B_excl_dynid, dtype=float)
        self.B[smap.B.lhs] = td[smap.B.rhs]
        self.B = np_.vstack((self.B, smap.dynid_B))

        self.C = np_.zeros(svec.shape_C_excl_dynid, dtype=float)
        self.C[smap.C.lhs] = tc[smap.C.rhs]
        self.C = np_.vstack((self.C, smap.dynid_C))

        self.D = np_.zeros(svec.shape_D_excl_dynid, dtype=float)
        self.D[smap.D.lhs] = td[smap.D.rhs]
        self.D = np_.vstack((self.D, smap.dynid_D))

        self.F = np_.zeros(svec.shape_F, dtype=float)
        self.F[smap.F.lhs] = td[smap.F.rhs]

        self.G = np_.zeros(svec.shape_G, dtype=float)
        self.G[smap.G.lhs] = td[smap.G.rhs]

        self.H = np_.zeros(svec.shape_H, dtype=float)
        self.H[smap.H.lhs] = tc[smap.H.rhs]

        self.J = np_.zeros(svec.shape_J, dtype=float)
        self.J[smap.J.lhs] = td[smap.J.rhs]

        return self





"""
First-order system simulators
"""

#[
from __future__ import annotations

from typing import (Self, TypeAlias, Literal, )
import numpy as np_

from ..fords import (solutions as sl_, descriptors as de_, )
from ..dataman import (databanks as db_, )
#]


def simulate_flat(
    solution: sl_.Solution,
    solution_vectors: de_.SolutionVectors,
    data: np_.ndarray,
    columns_to_run: list[int],
    deviation: bool,
    anticipate: bool,
):
    column_start = columns_to_run[0]
    column_slice = slice(column_start, column_start+len(columns_to_run))
    column_array = np_.array(range(column_start, column_start+len(columns_to_run)))

    vec = solution_vectors

    T = solution.T
    R = solution.R
    K = solution.K if not deviation else 0

    Z = solution.Z
    H = solution.H
    D = solution.D if not deviation else 0

    curr_state = np_.array([
        data[t.qid, column_start-1+t.shift] 
        for t in vec.transition_variables
    ], dtype=float).reshape(-1, 1)

    missing_initials = np_.isnan(curr_state)
    unnecessary_initials = ~np_.array(vec.initial_conditions).reshape(-1, 1)
    missing_unnecessary_initials = missing_initials & unnecessary_initials
    curr_state[missing_unnecessary_initials] = 0

    no_shift_state_to_slab_lhs = [
        t.qid
        for t in vec.transition_variables
        if t.shift == 0
    ]

    no_shift_state_to_slab_rhs = [
        j 
        for j, t in enumerate(vec.transition_variables)
        if t.shift == 0
    ]

    measurement_variables_to_slab = [t.qid for t in vec.measurement_variables]

    transition_shocks_in_slab = [t.qid for t in vec.transition_shocks]
    transition_shocks = data[transition_shocks_in_slab, :]

    measurement_shocks_in_slab = [t.qid for t in vec.measurement_shocks]
    measurement_shocks = data[[t.qid for t in vec.measurement_shocks], :]

    transition_shocks[np_.isnan(transition_shocks)] = 0
    measurement_shocks[np_.isnan(measurement_shocks)] = 0

    Rx = [R]
    forward = -1

    shock_column_incidence = list(np_.any(transition_shocks[:, column_array] != 0, axis=0))
    if any(shock_column_incidence):
        # Last shock at t+forward
        forward = len(shock_column_incidence) - shock_column_incidence[::-1].index(True) - 1
        if anticipate:
            Rk = solution.expand_square_solution(forward)
        else:
            Rk = [None]*forward
        Rx = Rx + (Rk if Rk is not None else [])

    shock_column_end = column_start + forward
    for t in column_array:
        shock_impact = sum( 
            Rx[k] @ transition_shocks[:, (s,)] if Rx[k] is not None else 0 
            for k, s in enumerate(range(t, shock_column_end+1))
        )
        curr_state = T @ curr_state + shock_impact + K
        data[no_shift_state_to_slab_lhs, t] = curr_state[no_shift_state_to_slab_rhs].flat

        y = Z @ curr_state + H @ measurement_shocks[:, (t,)] + D
        data[measurement_variables_to_slab, t] = y.flat

    data[transition_shocks_in_slab, :] = transition_shocks
    data[measurement_shocks_in_slab, :] = measurement_shocks

    return data




"""
"""


#[
import numpy as _np
from typing import (Callable, )
from collections.abc import (Iterable, )

from .. import (equations as _eq, )
from . import (abc as _abc, )
from ..evaluators import (steady as _es, )
#]


class SteadyEquator(_abc.EquatorMixin, ):
    """
    """
    #[
    def __init__(
        self,
        equations: _eq.Equations,
        t_zero: int,
        /,
        *,
        custom_functions: dict[str, Callable] | None = None,
    ) -> None:
        self._equations = tuple(equations, )
        self._create_equator_function(custom_functions, )
        self._populate_min_max_shifts()
        self._t_zero = t_zero

    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        ...
    #]


class FlatSteadyEquator(SteadyEquator, ):
    """
    """
    #[
    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        return self._func(steady_array, self._t_zero, steady_array, )
    #]


class NonflatSteadyEquator(SteadyEquator, ):
    """
    """
    #[
    def eval(
        self,
        steady_array: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        k = _es.NONFLAT_SHIFT
        return _np.hstack((
            self._func(steady_array, self._t_zero, steady_array, ),
            self._func(steady_array, self._t_zero+k, steady_array, ),
        ))
    #]





"""
Iris Pie
"""

from .dataman import *
from .dataman import __all__ as dataman_all

from .series import *
from .series import __all__ as series_all

from .models import *
from .models import __all__ as models_all

from .quantities import *
from .quantities import __all__ as quantities_all

from .equations import *
from .equations import __all__ as equations_all

__all__ = (
    []
    + dataman_all 
    + series_all 
    + models_all 
    + quantities_all
    + equations_all
)






import sys
sys.path.append("../../../src")

import irispie as _ip

db = _ip.Databank()
db.date = _ip.dater_from_sdmx_string(_ip.Frequency.QUARTERLY, "2020-Q1")
db.date_string = db.date.to_sdmx_string()

print(db)




"""
Mixins and protocols for equator classes
"""


#[
import numpy as _np
from typing import (Protocol, Callable, )
from collections.abc import (Iterable, )

from .. import (equations as _eq, quantities as _qu, )
from ..aldi import (adaptations as _aa, )
#]


class EquatorProtocol(Protocol):
    min_shift: int
    max_shift: int
    _equations: _eq.Equations
    _xtrings: Iterable[str]
    _func: Callable


class EquatorMixin:
    """
    """
    #[
    @property
    def equations_human(self: EquatorProtocol, /, ) -> tuple[str]:
        return tuple(eqn.human for eqn in self._equations)

    @property
    def num_equations(self: EquatorProtocol, /, ) -> int:
        """
        """
        return len(self._equations)

    def _create_equator_function(
        self: EquatorProtocol,
        /,
        function_context: dict | None = None,
    ) -> None:
        """
        """
        function_context = _aa.add_function_adaptations_to_custom_functions(function_context)
        function_context["_array"] = _np.array
        self._xtrings = [ eqn.remove_equation_ref_from_xtring() for eqn in self._equations ]
        func_string = " , ".join(self._xtrings)
        self._func = eval(_eq.EVALUATOR_PREAMBLE + f"_array([{func_string}], dtype=float)", function_context)

    def _populate_min_max_shifts(self: EquatorProtocol, /, ) -> None:
        """
        """
        self.min_shift = _eq.get_min_shift_from_equations(self._equations)
        self.max_shift = _eq.get_max_shift_from_equations(self._equations)
    #]






import sys
sys.path.append("../../../src")

import random as _ra

from irispie import *
from irispie.dataman import filters as _fi

x = Series.from_func(qq(2020,1)>>qq(2025,4), _ra.gauss, num_columns=2)





class ListException(Exception):
    def __init__(self, messages: list[str]) -> None:
        messages = "\n\n" + "\n".join([ "![modiphy] " + m for m in messages ]) + "\n"
        super().__init__(messages) 


class UndeclaredName(Exception):
    def __init__(self, errors: list[tuple[str, str]]) -> None:
        message = "".join([
            f"\n*** Undeclared or mistyped name '{e[0]}' in equation {e[1]}"
            for e in errors
        ])
        message = "\n" + message + "\n"
        super().__init__(message)

class UnknownName(Exception):
    def __init__(self, name):
        super().__init__(f"Name not found in Model object: '{name}'")




"""
Exporting data to CSV sheets
"""


#[
from __future__ import annotations

import csv as cs_
import numpy as np_
import dataclasses as dc_
import itertools as it_

from . import dates as _da
#]


class NotImplementedYet(Exception):
    pass


@dc_.dataclass
class _ExportBlockDescriptor:
    """
    """
    #[
    frequency: _da.Frequency | None = None
    names: Iterable[str] | None = None
    descriptors: Iterable[str] | None = None
    descriptor_row: bool | None = None
    num_series_columns: int | None = None
    dates: Iterable[Dater] | None = None
    data_array: np_.ndarray | None = None
    delimiter: str | None = None
    numeric_format: str | None = None
    nan_str: str | None = None

    def row_iterator(self):
        """
        """
        names = list(it_.chain.from_iterable( 
            [n] + ["*"]*(self.num_series_columns[i] - 1)
            for i, n in enumerate(self.names)
        ))
        yield [_get_frequency_mark(self.frequency)] + names
        if self.descriptor_row:
            descriptors = list(it_.chain.from_iterable( 
                [n] + ["*"]*(self.num_series_columns[i] - 1)
                for i, n in enumerate(self.descriptors)
            ))
            yield [""] + descriptors
        for date, data_row in zip(self.dates, self.data_array):
            yield [date] + [ x if not np_.isnan(x) else self.nan_str for x in data_row.tolist() ]
    #]


class DatabankExportMixin:
    """
    Databank mixin for exporting data to CSV sheets
    """
    #[
    def _to_sheet(
        self,
        file_name: str,
        /,
        descriptor_row: bool = False,
        range: Iterable[Dater] | None = None,
        frequency: _da.Frequency | None = None,
        delimiter: str = ",",
        numeric_format: str = "g",
        nan_str: str = "",
        csv_writer_settings: dict = {},
    ) -> Iterable[str]:
        """
        """
        if not frequency:
            raise NotImplementedYet("frequency=None")
        frequency = _da.Frequency(frequency)
        range = range if range else self._get_range_by_frequency(frequency)
        range = [ t for t in range ]
        num_data_rows = len(range)
        #
        names = self._get_series_names_by_frequency(frequency)
        descriptors = [ getattr(self, n)._description for n in names ]
        #
        num_series_columns = [ getattr(self, n).shape[1] for n in names ]
        data_array = np_.hstack([ getattr(self, n).get_data(range) for n in names ])
        export_block = _ExportBlockDescriptor(
            frequency, names, descriptors, descriptor_row,
            num_series_columns, range, data_array, 
            delimiter, numeric_format, nan_str,
        )
        with open(file_name, "w+") as fid:
            writer = cs_.writer(fid, delimiter=delimiter, **csv_writer_settings)
            for row in export_block.row_iterator():
                writer.writerow(row)
        #
        return names
    #]


def _get_frequency_mark(frequency, ):
    return "__" + frequency.name.lower() + "__"





"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np

from ..dataman import (databanks as _db, dataslabs as _ds, )
from ..fords import (simulators as _sr, )
from . import (variants as _va, )
#]


class _SimulatableProtocol(Protocol, ):
    """
    """
    #[
    num_variants: int
    _variants: Iterable[_va.Variant]

    def get_extended_range_from_base_range(): ...
    def get_ordered_names(): ...
    def get_solution_vectors(): ...
    #]


class Simulatable:
    """
    """
    #[
    def simulate(
        self: _SimulatableProtocol,
        in_databank: _db.Databank,
        base_range: Iterable[Dater],
        /,
        anticipate: bool = True,
        deviation: bool = False,
        prepend_input: bool = True,
    ) -> _db.Databank:
        """
        """
        ext_range, base_columns = self.get_extended_range_from_base_range(base_range)
        names = self.get_ordered_names()

        dataslabs = tuple(
            _ds.Dataslab.from_databank(in_databank, names, ext_range, column=i) 
            for i in range(self.num_variants)
        )

        for variant, dataslab in zip(self._variants, dataslabs):
            new_data = _sr.simulate_flat(
                variant.solution, self.get_solution_vectors(),
                _np.copy(dataslab.data), base_columns, deviation, anticipate,
            )
            dataslab.data = new_data
            dataslab.remove_columns(base_columns[-1] - len(ext_range) + 1)

        out_databank = _ds.multiple_to_databank(dataslabs)
        if prepend_input:
            under_databank = in_databank._copy()
            under_databank._clip(None, base_range[0]-1)
            out_databank._underlay(under_databank)

        out_databank = out_databank | self.get_parameters_stds()

        return out_databank
    #]





"""
Frequency conversion of time series
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )
import numpy as _np
import functools as _ft
import operator as _op

from ..dataman import dates as _da
#]


class ConversionMixin:
    """
    """
    #[
    def aggregate(
        self,
        new_freq: _da.Frequency,
        /,
        method: str = "mean",
        remove_missing: bool = False,
        select: list[int] | None = None,
    ) -> Self:
        """
        Aggregate time series to a lower frequency
        """
        method_func = _AGGREGATION_METHOD_RESOLUTION[method]
        #
        if new_freq == self.frequency:
            return self
        if new_freq > self.frequency or new_freq is _da.Frequency.UNKNOWN or self.frequency is _da.Frequency.UNKNOWN:
            raise ValueError(f"Cannot aggregate from {self.frequency} frequency to {new_freq} frequency")
        #
        new_dater_class = _da.DATER_CLASS_FROM_FREQUENCY_RESOLUTION[new_freq]
        #
        aggregate_within_data_func = _ft.partial(
            _aggregate_within_data,
            select,
            remove_missing,
            method_func,
        )
        #
        if self.frequency.is_regular:
            aggregate_func = _aggregate_regular_to_regular
        elif self.frequency is _da.Frequency.DAILY:
            aggregate_func = _aggregate_daily_to_regular
        #
        new_start_date, new_data = aggregate_func(self, new_dater_class, aggregate_within_data_func)
        return type(self).from_start_date_and_data(new_start_date, new_data)

    def disaggregate(
        self,
        new_freq: _da.Frequency,
        /,
        method: str = "flat",
    ) -> Self:
        """
        Disaggregate time series to a higher frequency
        """
        method_func = _DISAGGREGATION_METHOD_RESOLUTION[method]
        #
        if new_freq == self.frequency:
            return self
        if new_freq < self.frequency or new_freq is _da.Frequency.UNKNOWN or self.frequency is _da.Frequency.UNKNOWN:
            raise ValueError(f"Cannot aggregate from {self.frequency} frequency to {new_freq} frequency")
        #
        new_dater_class = _da.DATER_CLASS_FROM_FREQUENCY_RESOLUTION[new_freq]
        new_start_date, new_data = method_func(self, new_dater_class)
        return type(self).from_start_date_and_data(new_start_date, new_data)
    #]


def _aggregate_daily_to_regular(
    self,
    new_dater_class: type,
    aggregate_within_data_func: Callable,
) -> Self:
    """
    """
    #[
    start_date = self.start_date.to_start_of_year()
    end_date = self.end_date.to_end_of_year()
    all_self_data = self.get_data(_da.Ranger(start_date, end_date))
    start_year = self.start_date.get_year()
    end_year = self.end_date.get_year()
    new_start_date = new_dater_class.from_year_period(start_year, 1)
    new_end_date = new_dater_class.from_year_period(end_year, "end")
    get_slice_func = lambda t: slice(
        t.to_daily(position="start") - start_date,
        t.to_daily(position="end") - start_date + 1,
    )
    new_data = tuple(
        tuple(
            aggregate_within_data_func(column[get_slice_func(t)])
            for column in all_self_data.T
        )
        for t in _da.Ranger(new_start_date, new_end_date)
    )
    #
    new_data = _np.array(new_data, dtype=self.data_type)
    return new_start_date, new_data
    #]


def _aggregate_regular_to_regular(
    self,
    new_dater_class: type,
    aggregate_within_data_func: Callable,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    start_year = self.start_date.get_year()
    start_date = self.start_date.to_start_of_year()
    end_date = self.end_date.to_end_of_year()
    new_start_date = new_dater_class.from_year_period(start_year, 1)
    new_freq = new_dater_class.frequency
    self_data = self.get_data(_da.Ranger(start_date, end_date))
    factor = self.frequency.value // new_freq
    #
    # Loop over columns, create within-period data, and aggregate
    new_data_transposed = tuple(
        tuple(
            aggregate_within_data_func(within_data)
            for within_data in self_data_column.reshape((-1, factor))
        )
        for self_data_column in self_data.T
    )
    #
    new_data_transposed = _np.array(new_data_transposed, dtype=self.data_type)
    return new_start_date, new_data_transposed.T
    #]


def _disaggregate_flat(
    self,
    new_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    new_freq = new_dater_class.frequency
    new_start_date = new_dater_class.from_ymd(*self.start_date.to_ymd(position="start"))
    new_end_date = new_dater_class.from_ymd(*self.end_date.to_ymd(position="end"))
    new_num_periods = new_end_date - new_start_date + 1
    #
    get_slice_func = lambda t: slice(
        new_dater_class.from_ymd(*t.to_ymd(position="start")) - new_start_date,
        new_dater_class.from_ymd(*t.to_ymd(position="end")) - new_start_date + 1,
    )
    #
    new_data = _np.full((new_num_periods, self.num_columns), _np.nan, dtype=self.data_type)
    for t in _da.Ranger(self.start_date, self.end_date):
        new_data[get_slice_func(t), :] = self.get_data(t)
    return new_start_date, new_data
    #]


def _aggregate_within_data(
    select: list[int] | None,
    remove_missing: bool,
    method_func: Callable,
    within_data: _np.ndarray,
    /,
) -> _np.ndarray:
    """
    """
    #[
    if select is not None:
        within_data = within_data[select]
    if remove_missing:
        within_data = within_data[~_np.isnan(within_data)]
    return method_func(within_data) if within_data.size > 0 else _np.nan
    #]


_AGGREGATION_METHOD_RESOLUTION = {
    "mean": _np.mean,
    "sum": _np.sum,
    "first": _op.itemgetter(0),
    "last": _op.itemgetter(-1),
    "min": _np.min,
    "max": _np.max,
}


_DISAGGREGATION_METHOD_RESOLUTION = {
    "flat": _disaggregate_flat,
}




"""
Model equations
"""

#[
import enum as _en
import re as _re
import dataclasses as _dc
import itertools as _it
import operator as _op
from typing import (Self, TypeAlias, )
from collections.abc import (Iterable, )

from . import (quantities as _qu, incidence as _in, wrongdoings as _wd, )
#]


EVALUATOR_PREAMBLE = "lambda x, t, L: "
_EVALUATOR_FORMAT = EVALUATOR_PREAMBLE + "[{joined_xtrings}]"
X_REF_PATTERN = "{qid},t{shift:+g},{eid}"

_QUANTITY_NAME_PATTERN = _re.compile(r"\b([a-zA-Z]\w*)\b(\[[-+\d]+\])?(?!\()")
_EQUATION_REF = "..."
_REPLACE_NAME = "x[" + X_REF_PATTERN +"]"
_TRANSLATE_REF_TO_KEY = {
    "[": "['",
    "]": "]'",
}
_REPLACE_UKNOWN = "?"


ErrorLogType = list[tuple[str, str]]


class EquationKind(_en.Flag):
    """
    Classification of model equations
    """
    #[
    TRANSITION_EQUATION = _en.auto()
    MEASUREMENT_EQUATION = _en.auto()
    #]


__all__  = [
    "TRANSITION_EQUATION", "MEASUREMENT_EQUATION",
]
for n in __all__:
    exec(f"{n} = EquationKind.{n}")


@_dc.dataclass(slots=True, )
class Equation:
    """
    """
    id: int | None = None
    human: str | None = None
    kind: EquationKind | None = None
    description: str | None = None
    xtring: str | None = None
    incidence: _in.Tokens | None = None
    entry: int | None = None
    """
    """
    #[
    def finalize_from_human(self, name_to_id: dict[str, int]) -> tuple[Self, ErrorLogType]:
        self.xtring, self.incidence, error_log, *_ = _xtring_from_human(self.human, name_to_id)
        return self, error_log

    def replace_equation_ref_in_xtring(self, replacement):
        return self.xtring.replace(_EQUATION_REF, str(replacement))

    def remove_equation_ref_from_xtring(self):
        return self.xtring.replace(","+_EQUATION_REF, "")

    def set_id(self, qid: int) -> Self:
        self.id = qid
        return self

    def __hash__(self, /, ) -> int:
        return hash(self.__repr__)
    #]


Equations: TypeAlias = Iterable[Equation]


def generate_all_tokens_from_equations(equations: Equations) -> _in.Tokens:
    return _it.chain.from_iterable(eqn.incidence for eqn in equations)


def finalize_dynamic_equations(
    equations: Equations,
    name_to_id: dict[str, int],
) -> None:
    _finalize_equations_from_humans(equations, name_to_id)
    _replace_steady_ref(equations)


def finalize_steady_equations(
    equations: Equations,
    name_to_id: dict[str, int],
) -> None:
    _finalize_equations_from_humans(equations, name_to_id)
    _remove_steady_ref(equations)


def _replace_steady_ref(equations: Equations) -> None:
    STEADY_REF_PATTERN = _re.compile(r"&x\[([^,\]]+),[^\]]+\]")
    for eqn in equations:
        eqn.xtring = _re.sub(STEADY_REF_PATTERN, lambda match: "L["+match.group(1)+"]", eqn.xtring)


def _remove_steady_ref(equations: Equations) -> None:
    STEADY_REF_PATTERN = _re.compile(r"&x\b")
    for eqn in equations:
        eqn.xtring = _re.sub(STEADY_REF_PATTERN, "x", eqn.xtring)


def _finalize_equations_from_humans(
    equations: Equations,
    name_to_id: dict[str, int],
) -> None:
    """
    """
    #[
    error_log = []
    for eqn in equations:
        _, _error_log = eqn.finalize_from_human(name_to_id)
        error_log += _error_log
    if error_log:
        raise _wd.IrisPieError(
            ["Some names used in equations not declared"]
            + error_log
        )
    #]


def generate_names_from_human(human: str) -> Iterable[str]:
    """
    Generate all names from a single human string
    """
    return (f[0] for f in _QUANTITY_NAME_PATTERN.findall(human))


def generate_all_names(equations: Equations) -> list[str]:
    """
    Extract all names from a list of equations
    """
    return generate_names_from_human(" ".join(e.human for e in equations))


def create_name_to_qid_from_equations(equations: Equations) -> dict[str, int]:
    """
    """
    all_names = sorted(list(set(generate_all_names(equations))))
    return { name: qid for qid, name in enumerate(all_names) }


def create_equator_func_string(xtrings: str) -> str:
    """
    """
    return _EVALUATOR_FORMAT.format(joined_xtrings=" , ".join(xtrings))


def create_eid_to_wrt_tokens(
    equations: Equations,
    all_wrt_tokens: _in.Tokens,
) -> dict[int, _in.Tokens]:
    """
    """
    #[
    eid_to_wrt_tokens = {}
    for eqn in equations:
        eid_to_wrt_tokens[eqn.id] = _in.sort_tokens(
            wrt for wrt in all_wrt_tokens
            if wrt in eqn.incidence
        )
    return eid_to_wrt_tokens
    #]


def create_human_to_eid(
    equations: Equations,
    /,
) -> dict[str, int]:
    return { eqn.human: eqn.id for eqn in equations }


def _xtring_from_human( 
    human: str,
    name_to_id: dict[str, int],
) -> tuple[str, set[_in.Token], ErrorLogType, _in.Tokens]:
    """
    Convert human string to xtring and retrieve incidence tokens
    """
    #[
    tokens_list: list[_in.Token] = []
    error_log = []

    def _x_from_human(match: _re.Match) -> str:
        name = match.group(1)
        qid = name_to_id.get(name)
        if qid is not None:
            shift = _resolve_shift_str(match.group(2))
            tokens_list.append(_in.Token(qid, shift))
            return _REPLACE_NAME.format(
                qid=qid,
                shift=shift,
                eid=_EQUATION_REF,
            )
        else:
            error_log.append(name + " in " + human)
            tokens_list.append(_in.Token(None, None))
            return _REPLACE_UKNOWN

    xtring = _QUANTITY_NAME_PATTERN.sub(_x_from_human, human)
    xtring = _postprocess_xtring(xtring)
    return xtring, set(tokens_list), error_log, tokens_list
    #]


def _resolve_shift_str(shift_str: str) -> int:
    return int(shift_str.replace("[", "",).replace("]", "")) if shift_str is not None else 0


def _postprocess_xtring(equation: str) -> str:
    #[
    equation = equation.replace("^", "**")
    equation = equation.replace(" ", "")
    lhs_rhs = equation.split("=", maxsplit=1)
    if len(lhs_rhs)==2:
        equation = "-(" + lhs_rhs[0] + ")+" + lhs_rhs[1]
    return equation
    #]


def generate_all_eids(
    equations: Equations,
    /,
) -> Iterable[int]:
    return (eqn.id for eqn in equations)


def generate_eids_by_kind(
    equations: Equations,
    kind: EquationKind,
) -> Iterable[int]:
    return (eqn.id for eqn in equations if eqn.kind in kind)


def generate_equations_of_kind(
    equations: Equations,
    kind: EquationKind,
) -> Equations:
    return (eqn for eqn in equations if eqn.kind in kind)


def sort_equations(
    equations: Equations,
    /,
) -> Equations:
    return sorted(equations, key=_op.attrgetter("id"))


def get_min_shift_from_equations(
    equations: Equations,
    /,
) -> int:
    return _in.get_min_shift(generate_all_tokens_from_equations(equations))


def get_max_shift_from_equations(
    equations: Equations,
    /,
) -> int:
    return _in.get_max_shift(generate_all_tokens_from_equations(equations))


def validate_selection_of_equations(
    allowed_equations: Equations,
    custom_equations: Equations | None,
    /,
) -> tuple[Equations, Equations]:
    """
    """
    #[
    invalid_equations = list(set(custom_equations) - set(allowed_equations)) if custom_equations is not None else []
    custom_equations = list(custom_equations) if custom_equations is not None else list(allowed_equations)
    #]
    return custom_equations, invalid_equations


def lookup_eids_by_human_starts(
    equations: Equations,
    human_starts: Iterable[str],
    /,
) -> tuple[Equations, list[str]]:
    """
    """
    #[
    human_starts = list(human_starts)
    eids = [
        next((eqn.id for eqn in equations if eqn.human.startswith(hs)), None)
        for hs in human_starts
    ]
    valid_eids = [ i for i in eids if i is not None ]
    invalid_human_starts = [ h for h, i in zip(human_starts, eids) if i is None ]
    #]
    return valid_eids, invalid_human_starts


def create_incidence_matrix(
    equations: Equations,
    quantities: _qu.Quantities,
    /,
) -> None:
    """
    """
    num_equations = len(equations)
    num_quantities = len(quantities)
    incidence_matrix = _np.zeros((num_equations, num_quantities), dtype=bool)
    qids = list(_qu.generate_all_qids(quantities))
    qid_to_column = { qid: column for column, qid in enumerate(qids) }
    for row_index, eqn in enumerate(equations):
        column_indices = tuple(
            qid_to_column[tok.qid]
            for tok in eqn.incidence if tok.qid in qids
        )
        incidence_matrix[row_index, column_indices] = True
    return incidence_matrix





"""
"""


#[
import warnings
from numbers import Number
from typing import (Self, Literal, Callable, )
import numpy as _np
import operator as _op
from collections.abc import (Iterable, )

from .. import (quantities as _qu, )
#]


class Variant:
    """
    Container for parameter variant specific attributes of a model
    """
    #[
    __slots__ = (
        "levels", "changes", "solution",
    )

    _missing = _np.nan

    def __init__(
        self,
        quantities: _qu.Quantities, 
        /,
        **kwargs,
    ) -> None:
        self._initilize_values(quantities, )
        self.solution = None

    def _initilize_values(
        self,
        quantities: _qu.Quantities, 
        /,
    ) -> None:
        max_qid = _qu.get_max_qid(quantities, )
        size_array = max_qid + 1
        init_value = self._missing
        self.levels = _np.full((size_array,), init_value, dtype=float, )
        self.changes = _np.full((size_array,), init_value, dtype=float, )

    def update_values_from_dict(self, update: dict, /, ) -> None:
        self.levels = _update_from_dict(self.levels, update, _op.itemgetter(0), lambda x: x, )
        self.changes = _update_from_dict(self.changes, update, _op.itemgetter(1), lambda x: ..., )

    def update_levels_from_array(self, levels: _np.ndarray, qids: Iterable[int], /, ) -> None:
        self.levels = _update_from_array(self.levels, levels, qids, )

    def update_changes_from_array(self, changes: _np.ndarray, qids: Iterable[int], /, ) -> None:
        self.changes = _update_from_array(self.changes, changes, qids, )

    def retrieve_values(
        self,
        attr: Literal["levels"] | Literal["changes"],
        qids: Iterable[int]|None = None,
        /,
    ) -> _np.ndarray:
        values = _np.copy(getattr(self, attr)).reshape(-1, 1)
        return values[qids] if qids is not None else values

    def retrieve_maybelog_values_for_qids(
        self,
        qids: Iterable[int],
        qid_to_logly: dict[int, bool],
    ) -> tuple[_np.ndarray, _np.ndarray, ]:
        """
        """
        qids = list(qids)
        index_logly = list(_qu.generate_index_logly(qids, qid_to_logly, ))
        #
        # Extract initial guesses for levels and changes
        maybelog_levels = self.levels[qids].flatten()
        maybelog_changes = self.changes[qids].flatten()
        #
        # Logarithmize
        maybelog_levels[index_logly] = _np.log(maybelog_levels[index_logly])
        maybelog_changes[index_logly] = _np.log(maybelog_changes[index_logly])
        #
        return maybelog_levels, maybelog_changes

    def reset_changes(
        self,
        qid_to_logly: dict[int, bool],
        /,
    ) -> None:
        index_logly = self._generate_index_logly(qid_to_logly, )
        self.changes[:] = 0
        self.changes[index_logly] = 1

    def create_steady_array(
        self,
        qid_to_logly: dict[int, bool],
        /,
        num_columns: int = 1,
        shift_in_first_column: int = 0,
    ) -> _np.ndarray:
        """
        """
        levels = _np.copy(self.levels).reshape(-1, 1)
        changes = _np.copy(self.changes).reshape(-1, 1)
        #
        if num_columns==1 and shift_in_first_column==0:
            return levels
        #
        index_logly = list(self._generate_index_logly(qid_to_logly))
        #
        shift_vec = _np.array(range(shift_in_first_column, shift_in_first_column+num_columns))
        #
        warnings.filterwarnings(action="ignore", category=RuntimeWarning)
        levels[index_logly] = _np.log(levels[index_logly])
        changes[index_logly] = _np.log(changes[index_logly])
        warnings.filterwarnings(action="default", category=RuntimeWarning)
        #
        levels[_np.isnan(levels) | _np.isinf(levels)] = _np.nan
        changes[_np.isnan(changes) | _np.isinf(changes)] = 0
        #
        steady_array = levels + changes * shift_vec
        #
        warnings.filterwarnings(action="ignore", category=RuntimeWarning)
        steady_array[index_logly, :] = _np.exp(steady_array[index_logly, :])
        warnings.filterwarnings(action="default", category=RuntimeWarning)
        #
        return steady_array

    def _generate_index_logly(
        self,
        qid_to_logly: dict[int, bool],
        /,
    ) -> list[int]:
        """
        """
        qids = range(self.levels.shape[0])
        return _qu.generate_index_logly(qids, qid_to_logly, )

    def create_zero_array(
        self,
        qid_to_logly: dict[int, bool],
        /,
        num_columns: int = 1,
        shift_in_first_column: int = 0,
    ) -> _np.ndarray:
        """
        """
        levels = _np.copy(self.levels).reshape(-1, 1)
        inx_set_to_0 = [ qid for qid, logly in qid_to_logly.items() if logly is False ]
        inx_set_to_1 = [ qid for qid, logly in qid_to_logly.items() if logly is True ]
        levels[inx_set_to_0] = 0
        levels[inx_set_to_1] = 1
        return _np.tile(levels, (1, num_columns))
    #]


def _update_from_array(
    values: _np.ndarray,
    updated_values: _np.ndarray,
    qids: Iterable[int],
    /,
) -> _np.ndarray:
    """
    Update levels or changes from an array and a list of qids
    """
    #[
    if updated_values is not None:
        values[list(qids)] = updated_values.flat
    return values
    #]


def _update_from_dict(
    what_to_update: _np.ndarray,
    update: dict[int, Number|tuple],
    when_tuple: Callable,
    when_not_tuple: Callable,
    /,
) -> _np.ndarray:
    """
    Update levels or changes from a dictionary
    """
    #[
    for qid, value in update.items():
        new_value = when_tuple(value) if isinstance(value, tuple) else when_not_tuple(value)
        what_to_update[qid] = new_value if new_value is not ... else what_to_update[qid]
    return what_to_update
    #]





"""
"""


#[
from __future__ import annotations

import numpy as np_
#]


FUNCTION_ADAPTATIONS = [
    "log", "exp", "sqrt", "maximum", "minimum"
]


for n in FUNCTION_ADAPTATIONS:
    exec(
        f"def {n}(x, *args, **kwargs): "
        f"return x._{n}_(*args, **kwargs) if hasattr(x, '_{n}_') else np_.{n}(x, *args, **kwargs)"
    )


def add_function_adaptations_to_custom_functions(context: dict | None) -> dict:
    """
    """
    #[
    context = context if context else {}
    for n in FUNCTION_ADAPTATIONS:
        context[n] = globals()[n]
    return context
    #]




"""
"""


#[
import functools as _ft
import numpy as _np
import itertools as _it
from numbers import (Number, )
from collections.abc import (Iterable, Callable, )
from typing import (TypeAlias, Any, Literal, )

from .. import (equations as _eq, quantities as _qu, wrongdoings as _wd, )
from . import (variants as _va, flags as _mg, )
from ..fords import (steadiers as _fs, )
from ..evaluators import (steady as _es, )
#]


_SteadySolverReturn: TypeAlias = tuple[
    _np.ndarray, tuple[int], _np.ndarray, tuple[int],
]

_EquationSwitch: TypeAlias = Literal["steady"] | Literal["dynamic"]


class SteadyMixin:
    """
    """
    def steady(
        self,
        /,
        **kwargs, 
    ) -> list[dict]:
        """
        Calculate steady state for each Variant within this Model
        """
        model_flags = _mg.Flags.update_from_kwargs(self.get_flags(), **kwargs)
        solver = self._choose_steady_solver(model_flags.is_linear, model_flags.is_flat, )
        info = tuple({} for _ in self._variants)
        for i, v in enumerate(self._variants):
            levels, qids_levels, changes, qids_changes = solver(v, model_flags, )
            v.update_levels_from_array(levels, qids_levels, )
            v.update_changes_from_array(changes, qids_changes, )
        return info


    def _steady_linear(
        self, 
        variant: _va.Variant,
        model_flags: _mg.Flags,
        /,
        algorithm: Callable,
    ) -> _SteadySolverReturn:
        """
        """
        #
        # Calculate first-order system for steady equations for this variant
        system = self._systemize(
            variant,
            self._invariant._steady_descriptor,
            model_flags,
        )
        #
        #=======================================================================
        # Core algorithm: Calculate steady state for this variant
        Xi, Y, dXi, dY = algorithm(system)
        levels = _np.hstack(( Xi.flat, Y.flat )).flatten()
        changes = _np.hstack(( dXi.flat, dY.flat )).flatten()
        #=======================================================================
        #
        # Extract only tokens with zero shift
        tokens = tuple(_it.chain(
            self._invariant._steady_descriptor.system_vectors.transition_variables,
            self._invariant._steady_descriptor.system_vectors.measurement_variables,
        ))
        #
        # True for tokens with zero shift, e.g. [True, False, True, ... ]
        zero_shift_index = [ not t.shift for t in tokens ]
        #
        # Extract steady levels for quantities with zero shift
        levels = levels[zero_shift_index]
        changes = changes[zero_shift_index]
        qids = tuple(t.qid for t in _it.compress(tokens, zero_shift_index))
        #
        # Delogarithmize when needed
        qid_to_logly = self.create_qid_to_logly()
        levels = _apply_delog_on_vector(levels, qids, qid_to_logly)
        changes = _apply_delog_on_vector(changes, qids, qid_to_logly)
        #
        return levels, qids, changes, qids

    _steady_linear_flat = _ft.partialmethod(_steady_linear, algorithm=_fs.solve_steady_linear_flat)
    _steady_linear_nonflat = _ft.partialmethod(_steady_linear, algorithm=_fs.solve_steady_linear_nonflat)


    def _steady_nonlinear_flat(
        self,
        variant: _va.Variant,
        /,
    ) -> _SteadySolverReturn:
        """
        """
        return None, None, None, None


    def _steady_nonlinear_nonflat(
        self,
        variant: _va.Variant,
        /,
    ) -> _SteadySolverReturn:
        """
        """
        return None, None, None, None


    def _choose_steady_solver(
        self,
        is_linear: bool,
        is_flat: bool,
        /,
    ) -> Callable:
        """
        Choose steady solver depending on linear and flat flags
        """
        match (is_linear, is_flat):
            case (False, False):
                return self._steady_nonlinear_nonflat
            case (False, True):
                return self._steady_nonlinear_flat
            case (True, False):
                return self._steady_linear_nonflat
            case (True, True):
                return self._steady_linear_flat


    def check_steady(
        self,
        /,
        equation_switch: _EquationSwitch = "dynamic",
        when_fails: _wd.HOW = "error",
        tolerance: float = 1e-12,
    ) -> tuple[bool, tuple[dict, ...]]:
        """
        Verify currently assigned steady state in dynamic or steady equations for each Variant within this Model
        """
        qid_to_logly = self.create_qid_to_logly()
        equator = self._choose_plain_equator(equation_switch)
        steady_arrays = (
            v.create_steady_array(
                qid_to_logly,
                num_columns=equator.min_num_columns + 1,
                shift_in_first_column=equator.min_shift,
            ) for v in self._variants
        )
        # REFACTOR
        t_zero = -equator.min_shift
        dis = [ 
            _np.hstack((
                equator.eval(x, t_zero, x[:, t_zero]),
                equator.eval(x, t_zero+1, x[:, t_zero+1]),
            ))
            for x in steady_arrays
        ]
        # REFACTOR
        max_abs_dis = [ _np.max(_np.abs(d)) for d in dis ]
        status = [ d < tolerance for d in max_abs_dis ]
        all_status = all(status)
        if not all_status:
            message = "Invalid steady state"
            _wd.throw(when_fails, message)
        details = tuple(
            {"discrepancies": d, "max_abs_discrepancy": m, "is_valid": s}
            for d, m, s in zip(dis, max_abs_dis, status)
        )
        return all_status, details


def _apply_delog_on_vector(
    vector: _np.ndarray,
    qids: Iterable[int],
    qid_to_logly: dict[int, bool],
    /,
) -> _np.ndarray:
    """
    Delogarithmize the elements of numpy vector that have True log-status
    """
    logly_index = [ qid_to_logly[qid] for qid in qids ]
    vector = _np.copy(vector)
    if any(logly_index):
        vector[logly_index] = _np.exp(vector[logly_index])
    return vector




"""
Univariate time series filters
"""


#[
from __future__ import annotations

from numbers import (Number, )
from collections.abc import (Iterable, Callable, )
from types import (EllipsisType, )
from typing import (Self, )
import numpy as _np

from ..dataman import (dates as _da, )
#]


_AUTO_SMOOTH = {
    _da.Frequency.YEARLY: (10*1)**2,
    _da.Frequency.HALFYEARLY: (10*2)**2,
    _da.Frequency.QUARTERLY: (10*4)**2,
    _da.Frequency.MONTHLY: (10*12)**2,
    "default": (10*4)**2,
}


def _get_default_smooth(frequency, /, ):
    """
    Default smoothing parameter (lambda)
    """
    return _AUTO_SMOOTH.get(frequency, _AUTO_SMOOTH["default"])


class _ConstrainedHodrickPrescottFilter:
    """
    """
    #[
    def __init__(
        self,
        num_periods: int,
        smooth: Number,
        /,
        level_where: list[int] | None = None,
        change_where: list[int] | None = None,
        log: bool = False,
    ) -> None:
        self._num_periods = num_periods
        self._smooth = smooth
        self._log = log
        self._num_extra_rows = 0
        self._create_plain_filter_matrix()
        self._add_level_constraints(level_where, )
        self._add_change_constraints(change_where, )

    def filter_data(
        self,
        data: _np.ndarray,
        /,
        level_data: _np.ndarray | None = None,
        change_data: _np.ndarray | None = None,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        """
        extended_data = self._extend_data(data, level_data, change_data, )
        if self._log:
            extended_data = _np.log(extended_data, )
        enforced_zeros_where = _np.where(_np.isnan(data, ))
        extended_data[enforced_zeros_where] = 0
        trend_data = _np.linalg.solve(self._F, extended_data, )
        extended_data[enforced_zeros_where] = _np.nan
        gap_data = extended_data - trend_data
        if self._num_extra_rows > 0:
            trend_data = trend_data[:-self._num_extra_rows, :]
            gap_data = gap_data[:-self._num_extra_rows, :]
        if self._log:
            trend_data = _np.exp(trend_data, )
            gap_data = _np.exp(gap_data, )
        return trend_data, gap_data

    def _extend_data(
        self,
        data: _np.ndarray,
        level_data: _np.ndarray | None,
        change_data: _np.ndarray | None,
    ) -> _np.ndarray:
        """
        """
        extended_data = data.copy() if data.ndim > 1 else data.copy().reshape(-1, 1)
        tile_dim = (1, extended_data.shape[1], )
        if level_data is not None:
            extended_data = _np.vstack((extended_data, _np.tile(level_data, tile_dim), ), )
        if change_data is not None:
            extended_data = _np.vstack((extended_data, _np.tile(change_data, tile_dim), ), )
        return extended_data

    def _create_plain_filter_matrix(self, ) -> None:
        I = _np.eye(self._num_periods, dtype=float)
        K = _np.zeros((self._num_periods-2, self._num_periods), dtype=float)
        for i in range(self._num_periods-2):
            K[i,i] = 1
            K[i,i+2] = 1
        for i in range(self._num_periods-2):
            K[i,i+1] = -2
        self._F = I + self._smooth * (K.T @ K)

    def _add_level_constraints(self, level_where: list[int], /, ):
        if not level_where:
            return
        num_constraints = len(level_where)
        self._num_extra_rows += num_constraints
        extra_rows = _np.zeros((num_constraints, self._num_periods, ), dtype=float)
        extra_columns = _np.zeros((self._num_periods + num_constraints, num_constraints, ), dtype=float)
        for i, j in enumerate(level_where, ):
            extra_rows[i, j] = 1
            extra_columns[j, i] = 1
        self._F = _np.vstack((self._F, extra_rows, ))
        self._F = _np.hstack((self._F, extra_columns, ))

    def _add_change_constraints(self, change_where: list[int], /, ):
        if not change_where:
            return
        num_constraints = len(change_where)
        extra_rows = _np.zeros((num_constraints, self._F.shape[1], ), dtype=float, )
        extra_columns = _np.zeros((self._F.shape[0] + num_constraints, num_constraints, ), dtype=float, )
        for i, j in enumerate(change_where, ):
            extra_rows[i, [j-1, j]] = (-1, 1)
            extra_columns[[j-1, j], i] = (-1, 1)
        self._F = _np.vstack((self._F, extra_rows, ))
        self._F = _np.hstack((self._F, extra_columns, ))
        self._num_extra_rows += num_constraints
    #]


class HodrickPrescottMixin:
    """
    Constrained Hodrick-Prescott filter mixin for Series
    """
    #[
    def hpf(
        self,
        range: Iterable[_da.Dater] | EllipsisType = ...,
        smooth: Number | None = None,
        log: bool = False,
        /,
        level: Self | None = None,
        change: Self | None = None,
    ) -> tuple[Self, Self, ]:
        """
        Constrained Hodrick-Prescott filter
        """
        if smooth is None:
            smooth = _get_default_smooth(self.frequency)
        #
        encompassing_range = _da.get_encompassing_range(self, level, change, )
        range = [ t for t in encompassing_range.resolve(range) ]
        data = self.get_data(range)
        num_periods, num_columns = data.shape
        level_data, level_where = _prepare_constraints(level, range, )
        change_data, change_where = _prepare_constraints(change, range, )
        change_data, change_where = _remove_first_date_change(change_data, change_where, )
        #
        hp = _ConstrainedHodrickPrescottFilter(
            num_periods,
            smooth,
            level_where=level_where,
            change_where=change_where,
            log=log,
        )
        trend_data, gap_data = hp.filter_data(data, level_data=level_data, change_data=change_data, )
        #
        trend = type(self).from_dates_and_data(range, trend_data, )
        gap = type(self).from_dates_and_data(range, gap_data, )
        return trend, gap

    def hpf_trend(self, /, *args, **kwargs):
        return self.hpf(*args, **kwargs, )[0]

    def hpf_gap(self, /, *args, **kwargs):
        return self.hpf(*args, **kwargs, )[1]
    #]


def _prepare_constraints(
    constraint: Self | None,
    range: list[_da.Dater],
    /,
) -> tuple[_np.ndarray | None, list[int] | None, ]:
    #[
    if constraint is None:
        return None, None
    data = constraint.get_data(range, 0)
    where = list(_np.where(~_np.isnan(data.reshape(-1, ), ), )[0])
    data = data[where, :]
    return data, where
    #]


def _remove_first_date_change(
    change_data: _np.ndarray | None,
    change_where: list[int] | None,
) -> tuple[_np.ndarray, list[int], ]:
    #[
    if change_where and 0 in change_where:
        where = change_where.index(0)
        change_where.pop(where)
        change_data = _np.delete(change_data, where, axis=0)
        if not change_where:
            change_data = None
            change_where = None
    return change_data, change_where
    #]




"""
Model source
"""

#[
from __future__ import annotations

import re as _re
import collections
from typing import (TypeAlias, Literal, )
from collections.abc import (Iterable, )

from ..parsers import (preparser as _pp, model_source_parser as _pm, )
from .. import (equations as _eq, quantities as _qu, )
#]


QuantityInput: TypeAlias = tuple[str, str]
EquationInput: TypeAlias = tuple[str, tuple[str, str]]


STD_PREFIX = "std_"
_STD_DESCRIPTION = "(Std) "


LOGLY_VARIABLE = (
    _qu.QuantityKind.TRANSITION_VARIABLE
    | _qu.QuantityKind.MEASUREMENT_VARIABLE
    | _qu.QuantityKind.EXOGENOUS_VARIABLE
)


class ModelSource:
    """
    """
    #[
    __slots__ = (
        "quantities", "dynamic_equations", "steady_equations",
        "log_variables", "all_but", "context", "sealed",
    )
    def __init__(self, /) -> None:
        self.quantities = []
        self.dynamic_equations = []
        self.steady_equations = []
        self.log_variables = []
        self.all_but = []
        self.context = None
        self.sealed = False


    def seal(self, /) -> None:
        _check_unique_names(qty.human for qty in self.quantities)
        self._add_stds()
        self.quantities = _reorder_by_kind(self.quantities)
        self.dynamic_equations = _reorder_by_kind(self.dynamic_equations)
        self.steady_equations = _reorder_by_kind(self.steady_equations)
        self.quantities = _stamp_id(self.quantities)
        self.dynamic_equations = _stamp_id(self.dynamic_equations)
        self.steady_equations = _stamp_id(self.steady_equations)
        self._populate_logly()
        self.sealed = True


    @property
    def num_quantities(self, /) -> int:
        return len(self.quantities)


    @property
    def all_names(self, /) -> Iterable[str]:
        return [ qty.human for qty in self.quantities ]


    def add_parameters(self, names: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(names, _qu.QuantityKind.PARAMETER)


    def add_exogenous_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _qu.QuantityKind.EXOGENOUS_VARIABLE)


    def add_transition_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _qu.QuantityKind.TRANSITION_VARIABLE)


    def add_transition_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _qu.QuantityKind.TRANSITION_SHOCK)


    def add_transition_equations(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, _eq.EquationKind.TRANSITION_EQUATION)


    def add_measurement_variables(self, quantity_inputs: Iterable[EquationInput] | None) -> None:
        self._add_quantities(quantity_inputs, _qu.QuantityKind.MEASUREMENT_VARIABLE)


    def add_measurement_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _qu.QuantityKind.MEASUREMENT_SHOCK)


    def add_measurement_equations(self, equation_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_equations(equation_inputs, _eq.EquationKind.MEASUREMENT_EQUATION)


    def add_all_but(self, all_but_input: str | None) -> None:
        if not all_but_input:
            return
        self.all_but.append(all_but_input)


    def add_log_variables(self, log_variable_inputs: Iterable[str] | None, /, ) -> None:
        if not log_variable_inputs:
            return
        self.log_variables += log_variable_inputs


    def _add_quantities(self, quantity_inputs: Iterable[QuantityInput] | None, kind: _qu.QuantityKind) -> None:
        if not quantity_inputs:
            return
        offset = self.quantities[-1].entry + 1 if self.quantities else 0
        self.quantities = self.quantities + [
            _qu.Quantity(id=None, human=q[1].strip(), kind=kind, logly=None, description=q[0].strip(), entry=i)
            for i, q in enumerate(quantity_inputs, start=offset)
        ]


    def _add_equations(self, equation_inputs: Iterable[EquationInput] | None, kind: _eq.EquationKind, /) -> None:
        if not equation_inputs:
            return
        offset = self.dynamic_equations[-1].entry + 1 if self.dynamic_equations else 0
        self.dynamic_equations = self.dynamic_equations + [
            _eq.Equation(id=None, human=_handle_white_spaces(ein[1][0]), kind=kind, description=ein[0].strip(), entry=i)
            for i, ein in enumerate(equation_inputs, start=offset)
        ]
        self.steady_equations = self.steady_equations + [
            _eq.Equation(id=None, human=_handle_white_spaces(ein[1][1] if ein[1][1] else ein[1][0]), kind=kind, description=ein[0].strip(), entry=i)
            for i, ein in enumerate(equation_inputs, start=offset)
        ]


    def _add_stds(self) -> None:
        def _create_std_input_from_shock(shock):
            description = _STD_DESCRIPTION + (shock.description if shock.description else shock.human)
            human = STD_PREFIX + shock.human
            return(description, human)
        #
        transition_shocks = (q for q in self.quantities if q.kind in _qu.QuantityKind.TRANSITION_SHOCK)
        transition_std_inputs = (_create_std_input_from_shock(i) for i in transition_shocks)
        self._add_quantities(transition_std_inputs, _qu.QuantityKind.TRANSITION_STD)
        #
        measurement_shocks = (q for q in self.quantities if q.kind in _qu.QuantityKind.MEASUREMENT_SHOCK)
        measurement_std_inputs = (_create_std_input_from_shock(i) for i in measurement_shocks)
        self._add_quantities(measurement_std_inputs, _qu.QuantityKind.MEASUREMENT_STD)


    def _populate_logly(self, /) -> None:
        default_logly = self._resolve_default_logly()
        qid_to_logly = { 
            qty.id: default_logly if qty.human not in self.log_variables else not default_logly
            for qty in self.quantities
            if qty.kind in LOGLY_VARIABLE
        }
        self.quantities = [
            qty.set_logly(qid_to_logly.get(qty.id, None))
            for qty in self.quantities
        ]


    def _is_logly_consistent(self, /) -> bool:
        return all(a=="all-but" for a in self.all_but) or all(a=="" for a in self.all_but) if self.all_but else True


    def _resolve_default_logly(self, /) -> bool:
        if not self._is_logly_consistent():
            raise Exception("Inconsistent use of !all-but in !log-variables")
        return self.all_but.pop()=="all-but" if self.all_but else False


    @classmethod
    def from_lists(
        cls,
        /,
        transition_variables: Iterable[QuantityInput],
        transition_equations: Iterable[EquationInput], 
        transition_shocks: Iterable[QuantityInput] | None = None,
        measurement_variables: Iterable[QuantityInput] | None = None,
        measurement_equations: Iterable[EquationInput] | None = None,
        measurement_shocks: Iterable[QuantityInput] | None = None,
        parameters: Iterable[QuantityInput] | None = None,
        exogenous_variables: Iterable[QuantityInput] | None = None,
        log_variables: Iterable[str] | None = None,
        seal: bool = True,
    ) -> Self:
        """
        """
        self = ModelSource()
        self.add_transition_variables(transition_variables)
        self.add_transition_equations(transition_equations)
        self.add_transition_shocks(transition_shocks)
        self.add_measurement_variables(measurement_variables)
        self.add_measurement_equations(measurement_equations)
        self.add_measurement_shocks(measurement_shocks)
        self.add_parameters(parameters)
        self.add_exogenous_variables(exogenous_variables)
        self.add_log_variables(log_variables)
        if seal: self.seal()
        return self

    @classmethod
    def from_string(
        cls,
        source_string: str,
        /,
        context: dict | None = None,
        save_preparsed: str = "",
        seal: bool = True,
    ) -> tuple[Self, dict]:
        """
        """
        preparsed_string, preparser_info = _pp.from_string(
            source_string, context, save_preparsed=save_preparsed, 
        )
        parsed_content = _pm.from_string(preparsed_string)
        #
        self = ModelSource()
        self.add_transition_variables(parsed_content.get("transition-variables"))
        self.add_transition_shocks(parsed_content.get("transition-shocks"))
        self.add_transition_equations(parsed_content.get("transition-equations"))
        self.add_measurement_variables(parsed_content.get("measurement-variables"))
        self.add_measurement_shocks(parsed_content.get("measurement-shocks"))
        self.add_measurement_equations(parsed_content.get("measurement-equations"))
        self.add_parameters(parsed_content.get("parameters"))
        self.add_exogenous_variables(parsed_content.get("exogenous-variables"))
        self.add_log_variables(parsed_content.get("log-variables"))
        for i in parsed_content.get("all-but", []):
            self.add_all_but(i)
        self.context = preparser_info["context"]
        #
        if seal: 
            self.seal()
        #
        return self, preparser_info
    #]


def _check_unique_names(names: Iterable[str], /) -> None:
    """
    """
    #[
    name_counter = collections.Counter(names)
    if any(c>1 for c in name_counter.values()):
        duplicates = ( n for n, c in name_counter.items() if c>1 )
        raise Exception("Duplicate names " + ", ".join(duplicates))
    #]


def _handle_white_spaces(x: str, /) -> str:
    return _re.sub(r"[\s\n\r]+", "", x)


def _reorder_by_kind(items: Iterable, /) -> Iterable:
    return sorted(items, key=lambda x: (x.kind.value, x.entry))


def _stamp_id(items: Iterable, /) -> Iterable:
    return [ i.set_id(_id) for _id, i in enumerate(items) ]






#[
from __future__ import annotations

import numpy as np_
from typing import (Self, )
from collections.abc import (Iterable, )

from ..series import facade as _sf
from . import databanks as db_
from . import dates as da_
#]


class Dataslab:
    """
    """
    #[
    data: np_.ndarray | None = None
    row_names: Iterable[str] | None = None,
    missing_names: Iterable[str] | None = None,
    column_dates: Iterable[Dater] | None = None

    def to_databank(self, *args, **kwargs, ) -> da_.Databank:
        """
        """
        return multiple_to_databank((self,), *args, **kwargs, )

    def add_from_databank(
        self,
        in_databank: db_.Databank,
        names: Iterable[str],
        ext_range: Ranger,
        /,
        column: int = 0,
    ) -> Self:
        """
        Add data from a databank to this dataslab
        """
        missing_names = [
            n for n in names
            if not hasattr(in_databank, n)
        ]
        #
        num_periods = len(ext_range)
        nan_row = np_.full((1, num_periods), np_.nan, dtype=float)
        generate_data = tuple(
            _extract_data_from_record(getattr(in_databank, n), ext_range, column)
            if n not in missing_names else nan_row
            for n in names
        )
        #
        self.data = np_.vstack(generate_data)
        self.row_names = tuple(names)
        self.column_dates = tuple(ext_range)
        self.missing_names = missing_names
        return self

    def remove_columns(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove > 0:
            self.data = self.data[:, remove:]
            self.column_dates = self.column_dates[remove:]
        elif remove < 0:
            self.data = self.data[:, :remove]
            self.column_dates = self.column_dates[:remove]

    @classmethod
    def from_databank(
        cls,
        *args,
        **kwargs,
    ) -> Self:
        """
        Create a new dataslab from a databank
        """
        self = cls();
        self.add_from_databank(*args, **kwargs)
        return self
    #]


def _extract_data_from_record(record, ext_range, column, /, ):
    """
    """
    return (
        record.get_data_column(ext_range, column).reshape(1, -1) 
        if hasattr(record, "get_data")
        else np_.full((1, len(ext_range)), float(record), dtype=float)
    )


def multiple_to_databank(
    selves,
    /,
    target_databank: db_.Databank | None = None,
) -> db_.Databank:
    """
    Add data from a dataslab to a new or existing databank
    """
    out_databank = target_databank if target_databank is not None else db_.Databank()
    self = selves[0]
    num_columns = len(selves)
    for row, n in enumerate(self.row_names):
        data = np_.hstack(tuple(ds.data[(row,), :].T for ds in selves))
        x = _sf.Series(num_columns=num_columns)
        x.set_data(self.column_dates, data)
        setattr(out_databank, n, x)
    return out_databank

